# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Label encoder for YOLOV8. This uses the TOOD Task Aligned Assigner approach.
See https://arxiv.org/abs/2108.07755 for more info, as well as a reference
implementation at https://github.com/fcjian/TOOD/blob/master/mmdet/core/bbox/assigners/task_aligned_assigner.py
"""  # noqa: E501

import tensorflow as tf

from keras_cv.src import bounding_box
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.bounding_box.iou import compute_ciou


def is_anchor_center_within_box(anchors, gt_bboxes):
    return ops.all(
        ops.logical_and(
            gt_bboxes[:, :, None, :2] < anchors,
            gt_bboxes[:, :, None, 2:] > anchors,
        ),
        axis=-1,
    )


@keras_cv_export("keras_cv.models.yolov8.LabelEncoder")
class YOLOV8LabelEncoder(keras.layers.Layer):
    """
    Encodes ground truth boxes to target boxes and class labels for training a
    YOLOV8 model. This is an implementation of the Task-aligned sample
    assignment scheme proposed in https://arxiv.org/abs/2108.07755.

    Args:
        num_classes: integer, the number of classes in the training dataset
        max_anchor_matches: optional integer, the maximum number of anchors to
            match with any given ground truth box. For example, when the default
            10 is used, the 10 candidate anchor points with the highest
            alignment score are matched with a ground truth box. If less than 10
            candidate anchors exist, all candidates will be matched to the box.
        alpha: float, a parameter to control the influence of class predictions
            on the alignment score of an anchor box. This is the alpha parameter
            in equation 9 of https://arxiv.org/pdf/2108.07755.pdf.
        beta: float, a parameter to control the influence of box IOUs on the
            alignment score of an anchor box. This is the beta parameter in
            equation 9 of https://arxiv.org/pdf/2108.07755.pdf.
        epsilon: float, a small number used for numerical stability in division
            (to avoid diving by zero), and used as a threshold to eliminate very
            small matches based on alignment scores of approximately zero.
    """

    def __init__(
        self,
        num_classes,
        max_anchor_matches=10,
        alpha=0.5,
        beta=6.0,
        epsilon=1e-9,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.max_anchor_matches = max_anchor_matches
        self.num_classes = num_classes
        self.alpha = alpha
        self.beta = beta
        self.epsilon = epsilon

    def assign(
        self, scores, decode_bboxes, anchors, gt_labels, gt_bboxes, gt_mask
    ):
        """Assigns ground-truth boxes to anchors.

        Uses the task-aligned assignment strategy for matching ground truth
        and anchor boxes based on prediction scores and IoU.
        """
        num_anchors = anchors.shape[0]

        # Box scores are the predicted scores for each anchor, ground truth box
        # pair. Only the predicted score for the class of the GT box is included
        # Shape: (B, num_gt_boxes, num_anchors) (after transpose)
        bbox_scores = ops.take_along_axis(
            scores,
            ops.cast(ops.maximum(gt_labels[:, None, :], 0), "int32"),
            axis=-1,
        )
        bbox_scores = ops.transpose(bbox_scores, (0, 2, 1))

        # Overlaps are the IoUs of each predicted box and each GT box.
        # Shape: (B, num_gt_boxes, num_anchors)
        overlaps = compute_ciou(
            ops.expand_dims(gt_bboxes, axis=2),
            ops.expand_dims(decode_bboxes, axis=1),
            bounding_box_format="xyxy",
        )

        # Alignment metrics are a combination of box scores and overlaps, per
        # the task-aligned-assignment formula.
        # Metrics are forced to 0 for boxes which have been masked in the GT
        # input (e.g. due to padding)
        alignment_metrics = ops.power(bbox_scores, self.alpha) * ops.power(
            overlaps, self.beta
        )
        alignment_metrics = ops.where(gt_mask, alignment_metrics, 0)

        # Only anchors which are inside of relevant GT boxes are considered
        # for assignment.
        # This is a boolean tensor of shape (B, num_gt_boxes, num_anchors)
        matching_anchors_in_gt_boxes = is_anchor_center_within_box(
            anchors, gt_bboxes
        )
        alignment_metrics = ops.where(
            matching_anchors_in_gt_boxes, alignment_metrics, 0
        )

        # The top-k highest alignment metrics are used to select K candidate
        # anchors for each GT box.
        candidate_metrics, candidate_idxs = ops.top_k(
            alignment_metrics, self.max_anchor_matches
        )
        candidate_idxs = ops.where(candidate_metrics > 0, candidate_idxs, -1)

        # We now compute a dense grid of anchors and GT boxes. This is useful
        # for picking a GT box when an anchor matches to 2, as well as returning
        # to a dense format for a mask of which anchors have been matched.
        anchors_matched_gt_box = ops.zeros_like(overlaps)
        for k in range(self.max_anchor_matches):
            anchors_matched_gt_box += ops.one_hot(
                candidate_idxs[:, :, k], num_anchors
            )

        # We zero-out the overlap for anchor, GT box pairs which don't match.
        overlaps *= anchors_matched_gt_box
        # In cases where one anchor matches to 2 GT boxes, we pick the GT box
        # with the highest overlap as a max.
        gt_box_matches_per_anchor = ops.argmax(overlaps, axis=1)
        gt_box_matches_per_anchor_mask = ops.max(overlaps, axis=1) > 0
        # TODO(ianstenbit): Once ops.take_along_axis supports -1 in Torch,
        # replace gt_box_matches_per_anchor with
        # ops.where(
        #     ops.max(overlaps, axis=1) > 0, ops.argmax(overlaps, axis=1), -1
        # )
        # and get rid of the manual masking
        gt_box_matches_per_anchor = ops.cast(gt_box_matches_per_anchor, "int32")

        # We select the GT boxes and labels that correspond to anchor matches.
        bbox_labels = ops.take_along_axis(
            gt_bboxes, gt_box_matches_per_anchor[:, :, None], axis=1
        )
        bbox_labels = ops.where(
            gt_box_matches_per_anchor_mask[:, :, None], bbox_labels, -1
        )
        class_labels = ops.take_along_axis(
            gt_labels, gt_box_matches_per_anchor, axis=1
        )
        class_labels = ops.where(
            gt_box_matches_per_anchor_mask, class_labels, -1
        )

        class_labels = ops.one_hot(
            ops.cast(class_labels, "int32"), self.num_classes
        )

        # Finally, we normalize an anchor's class labels based on the relative
        # strength of the anchors match with the corresponding GT box.
        alignment_metrics *= anchors_matched_gt_box
        max_alignment_per_gt_box = ops.max(
            alignment_metrics, axis=-1, keepdims=True
        )
        max_overlap_per_gt_box = ops.max(overlaps, axis=-1, keepdims=True)

        normalized_alignment_metrics = ops.max(
            alignment_metrics
            * max_overlap_per_gt_box
            / (max_alignment_per_gt_box + self.epsilon),
            axis=-2,
        )
        class_labels *= normalized_alignment_metrics[:, :, None]

        # On TF backend, the final "4" becomes a dynamic shape so we include
        # this to force it to a static shape of 4. This does not actually
        # reshape the Tensor.
        bbox_labels = ops.reshape(bbox_labels, (-1, num_anchors, 4))
        return (
            ops.stop_gradient(bbox_labels),
            ops.stop_gradient(class_labels),
            ops.stop_gradient(
                ops.cast(gt_box_matches_per_anchor > -1, "float32")
            ),
        )

    def call(
        self, scores, decode_bboxes, anchors, gt_labels, gt_bboxes, gt_mask
    ):
        """Computes target boxes and classes for anchors.

        Args:
            scores: a Float Tensor of shape (batch_size, num_anchors,
                num_classes) representing predicted class scores for each
                anchor.
            decode_bboxes: a Float Tensor of shape (batch_size, num_anchors, 4)
                representing predicted boxes for each anchor.
            anchors: a Float Tensor of shape (batch_size, num_anchors, 2)
                representing the xy coordinates of the center of each anchor.
            gt_labels: a Float Tensor of shape (batch_size, num_gt_boxes)
                representing the classes of ground truth boxes.
            gt_bboxes: a Float Tensor of shape (batch_size, num_gt_boxes, 4)
                representing the ground truth bounding boxes in xyxy format.
            gt_mask: A Boolean Tensor of shape (batch_size, num_gt_boxes)
                representing whether a box in `gt_bboxes` is a real box or a
                non-box that exists due to padding.

        Returns:
            A tuple of the following:
                - A Float Tensor of shape (batch_size, num_anchors, 4)
                    representing box targets for the model.
                - A Float Tensor of shape (batch_size, num_anchors, num_classes)
                    representing class targets for the model.
                - A Boolean Tensor of shape (batch_size, num_anchors)
                    representing whether each anchor was a match with a ground
                    truth box. Anchors that didn't match with a ground truth
                    box should be excluded from both class and box losses.
        """
        if isinstance(gt_bboxes, tf.RaggedTensor):
            dense_bounding_boxes = bounding_box.to_dense(
                {"boxes": gt_bboxes, "classes": gt_labels},
            )
            gt_bboxes = dense_bounding_boxes["boxes"]
            gt_labels = dense_bounding_boxes["classes"]

        if isinstance(gt_mask, tf.RaggedTensor):
            gt_mask = gt_mask.to_tensor()

        max_num_boxes = ops.shape(gt_bboxes)[1]

        # If there are no GT boxes in the batch, we short-circuit and return
        # empty targets to avoid NaNs.
        return ops.cond(
            ops.array(max_num_boxes > 0),
            lambda: self.assign(
                scores, decode_bboxes, anchors, gt_labels, gt_bboxes, gt_mask
            ),
            lambda: (
                ops.zeros_like(decode_bboxes),
                ops.zeros_like(scores),
                ops.zeros_like(scores[..., 0]),
            ),
        )

    def count_params(self):
        # The label encoder has no weights, so we short-circuit the weight
        # counting to avoid having to `build` this layer unnecessarily.
        return 0

    def get_config(self):
        config = {
            "max_anchor_matches": self.max_anchor_matches,
            "num_classes": self.num_classes,
            "alpha": self.alpha,
            "beta": self.beta,
            "epsilon": self.epsilon,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
formats.py contains axis information for each supported format.
"""


class CENTER_XYZ_DXDYDZ_PHI:
    """CENTER_XYZ_DXDYDZ_PHI contains axis indices for the CENTER_XYZ_DXDYDZ_PHI
    format.

    CENTER_XYZ_DXDYDZ_PHI is a 3D box format that supports vertical boxes with a
    heading rotated around the Z axis.

    The CENTER_XYZ_DXDYDZ_PHI format consists of the following required indices:

    - X: X coordinate of the center of the bounding box
    - Y: Y coordinate of the center of the bounding box
    - Z: Z coordinate of the center of the bounding box
    - DX: size of the bounding box on the x-axis
    - DY: size of the bounding box on the y-axis
    - DZ: size of the bounding box on the z-axis
    - PHI: the rotation of the box with respect to the z axis, in radians

    and the following optional indices, used in some KerasCV components:

    - CLASS: class of the object contained in the bounding box
    """

    X = 0
    Y = 1
    Z = 2
    DX = 3
    DY = 4
    DZ = 5
    PHI = 6
    CLASS = 7



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops


@keras_cv_export("keras_cv.losses.FocalLoss")
class FocalLoss(keras.losses.Loss):
    """Implements Focal loss

    Focal loss is a modified cross-entropy designed to perform better with
    class imbalance. For this reason, it's commonly used with object detectors.

    Args:
        alpha: a float value between 0 and 1 representing a weighting factor
            used to deal with class imbalance. Positive classes and negative
            classes have alpha and (1 - alpha) as their weighting factors
            respectively. Defaults to 0.25.
        gamma: a positive float value representing the tunable focusing
            parameter, defaults to 2.
        from_logits: Whether `y_pred` is expected to be a logits tensor. By
            default, `y_pred` is assumed to encode a probability distribution.
            Default to `False`.
        label_smoothing: Float in `[0, 1]`. If higher than 0 then smooth the
            labels by squeezing them towards `0.5`, i.e., using
            `1. - 0.5 * label_smoothing` for the target class and
            `0.5 * label_smoothing` for the non-target class.

    References:
        - [Focal Loss paper](https://arxiv.org/abs/1708.02002)

    Example:
    ```python
    y_true = np.random.uniform(size=[10], low=0, high=4)
    y_pred = np.random.uniform(size=[10], low=0, high=4)
    loss = FocalLoss()
    loss(y_true, y_pred)
    ```
    Usage with the `compile()` API:
    ```python
    model.compile(optimizer='adam', loss=keras_cv.losses.FocalLoss())
    ```
    """

    def __init__(
        self,
        alpha=0.25,
        gamma=2,
        from_logits=False,
        label_smoothing=0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.alpha = float(alpha)
        self.gamma = float(gamma)
        self.from_logits = from_logits
        self.label_smoothing = label_smoothing

    def _smooth_labels(self, y_true):
        return (
            y_true * (1.0 - self.label_smoothing) + 0.5 * self.label_smoothing
        )

    def call(self, y_true, y_pred):
        y_pred = ops.convert_to_tensor(y_pred)
        y_true = ops.cast(y_true, y_pred.dtype)

        if self.label_smoothing:
            y_true = self._smooth_labels(y_true)

        if self.from_logits:
            y_pred = ops.sigmoid(y_pred)

        cross_entropy = ops.binary_crossentropy(y_true, y_pred)

        alpha = ops.where(
            ops.equal(y_true, 1.0), self.alpha, (1.0 - self.alpha)
        )
        pt = y_true * y_pred + (1.0 - y_true) * (1.0 - y_pred)
        loss = (
            alpha
            * ops.cast(ops.power(1.0 - pt, self.gamma), alpha.dtype)
            * ops.cast(cross_entropy, alpha.dtype)
        )
        # In most losses you mean over the final axis to achieve a scalar
        # Focal loss however is a special case in that it is meant to focus on
        # a small number of hard examples in a batch. Most of the time this
        # comes in the form of thousands of background class boxes and a few
        # positive boxes.
        # If you mean over the final axis you will get a number close to 0,
        # which will encourage your model to exclusively predict background
        # class boxes.
        return ops.sum(loss, axis=-1)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "alpha": self.alpha,
                "gamma": self.gamma,
                "from_logits": self.from_logits,
                "label_smoothing": self.label_smoothing,
            }
        )
        return config



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.backend import keras
from keras_cv.src.models.stable_diffusion.padded_conv2d import PaddedConv2D


class ResnetBlock(keras.layers.Layer):
    def __init__(self, output_dim, **kwargs):
        super().__init__(**kwargs)
        self.output_dim = output_dim
        self.norm1 = keras.layers.GroupNormalization(epsilon=1e-5)
        self.conv1 = PaddedConv2D(output_dim, 3, padding=1)
        self.norm2 = keras.layers.GroupNormalization(epsilon=1e-5)
        self.conv2 = PaddedConv2D(output_dim, 3, padding=1)

    def build(self, input_shape):
        if input_shape[-1] != self.output_dim:
            self.residual_projection = PaddedConv2D(self.output_dim, 1)
        else:
            self.residual_projection = lambda x: x

    def call(self, inputs):
        x = self.conv1(keras.activations.swish(self.norm1(inputs)))
        x = self.conv2(keras.activations.swish(self.norm2(x)))
        return x + self.residual_projection(inputs)



# Copyright 2024 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.layers import DropPath


def window_partition(x, window_size):
    """Partitions a video tensor into non-overlapping windows of a specified size.

    Args:
        x: A tensor with shape (B, D, H, W, C), where:
            - B: Batch size
            - D: Number of frames (depth) in the video
            - H: Height of the video frames
            - W: Width of the video frames
            - C: Number of channels in the video (e.g., RGB for color)
        window_size: A tuple of ints of size 3 representing the window size
            along each dimension (depth, height, width).

    Returns:
        A tensor with shape (num_windows * B, window_size[0], window_size[1], window_size[2], C),
        where each window from the video is a sub-tensor containing the specified
        number of frames and the corresponding spatial window.
    """  # noqa: E501

    input_shape = ops.shape(x)
    batch_size, depth, height, width, channel = (
        input_shape[0],
        input_shape[1],
        input_shape[2],
        input_shape[3],
        input_shape[4],
    )

    x = ops.reshape(
        x,
        [
            batch_size,
            depth // window_size[0],
            window_size[0],
            height // window_size[1],
            window_size[1],
            width // window_size[2],
            window_size[2],
            channel,
        ],
    )

    x = ops.transpose(x, [0, 1, 3, 5, 2, 4, 6, 7])
    windows = ops.reshape(
        x, [-1, window_size[0] * window_size[1] * window_size[2], channel]
    )

    return windows


def window_reverse(windows, window_size, batch_size, depth, height, width):
    """Reconstructs the original video tensor from its partitioned windows.

    This function assumes the windows were created using the `window_partition` function
    with the same `window_size`.

    Args:
        windows: A tensor with shape (num_windows * batch_size, window_size[0],
            window_size[1], window_size[2], channels), where:
            - num_windows: Number of windows created during partitioning
            - channels: Number of channels in the video (same as in `window_partition`)
        window_size: A tuple of ints of size 3 representing the window size used
            during partitioning (same as in `window_partition`).
        batch_size: Batch size of the original video tensor (same as in `window_partition`).
        depth: Number of frames (depth) in the original video tensor (same as in `window_partition`).
        height: Height of the video frames in the original tensor (same as in `window_partition`).
        width: Width of the video frames in the original tensor (same as in `window_partition`).

    Returns:
        A tensor with shape (batch_size, depth, height, width, channels), representing the
        original video reconstructed from the provided windows.
    """  # noqa: E501
    x = ops.reshape(
        windows,
        [
            batch_size,
            depth // window_size[0],
            height // window_size[1],
            width // window_size[2],
            window_size[0],
            window_size[1],
            window_size[2],
            -1,
        ],
    )
    x = ops.transpose(x, [0, 1, 4, 2, 5, 3, 6, 7])
    x = ops.reshape(x, [batch_size, depth, height, width, -1])
    return x


def get_window_size(x_size, window_size, shift_size=None):
    """Computes the appropriate window size and potentially shift size for Swin Transformer.

    This function implements the logic from the Swin Transformer paper by Ze Liu et al.
    (https://arxiv.org/abs/2103.14030) to determine suitable window sizes
    based on the input size and the provided base window size.

    Args:
        x_size: A tuple of ints of size 3 representing the input size (depth, height, width)
            of the data (e.g., video).
        window_size: A tuple of ints of size 3 representing the base window size
            (depth, height, width) to use for partitioning.
        shift_size: A tuple of ints of size 3 (optional) representing the window
            shifting size (depth, height, width) for shifted window processing
            used in Swin Transformer. If not provided, only window size is computed.

    Returns:
        A tuple or a pair of tuples:
            - If `shift_size` is None, returns a single tuple representing the adjusted
            window size that may be smaller than the provided `window_size` to ensure
            it doesn't exceed the input size along any dimension.
            - If `shift_size` is provided, returns a pair of tuples. The first tuple
            represents the adjusted window size, and the second tuple represents the
            adjusted shift size. The adjustments ensure both window size and shift size
            do not exceed the corresponding dimensions in the input data.
    """  # noqa: E501

    use_window_size = list(window_size)

    if shift_size is not None:
        use_shift_size = list(shift_size)

    for i in range(len(x_size)):
        if x_size[i] <= window_size[i]:
            use_window_size[i] = x_size[i]
            if shift_size is not None:
                use_shift_size[i] = 0

    if shift_size is None:
        return tuple(use_window_size)
    else:
        return tuple(use_window_size), tuple(use_shift_size)


def compute_mask(depth, height, width, window_size, shift_size):
    """Computes an attention mask for a sliding window self-attention mechanism
    used in Video Swin Transformer.

    This function creates a mask to indicate which windows can attend to each other
    during the self-attention operation. It considers non-overlapping and potentially
    shifted windows based on the provided window size and shift size.

    Args:
        depth (int): Depth (number of frames) of the input video.
        height (int): Height of the video frames.
        width (int): Width of the video frames.
        window_size (tuple[int]): Size of the sliding window in each dimension
            (depth, height, width).
        shift_size (tuple[int]): Size of the shifting step in each dimension
            (depth, height, width).

    Returns:
        A tensor of shape (batch_size, num_windows, num_windows), where:
            - batch_size: Assumed to be 1 in this function.
            - num_windows: Total number of windows covering the entire input based on
                the formula:
                    (depth - window_size[0]) // shift_size[0] + 1) *
                    (height - window_size[1]) // shift_size[1] + 1) *
                    (width - window_size[2]) // shift_size[2] + 1)
        Each element (attn_mask[i, j]) represents the attention weight between
        window i and window j. A value of -100.0 indicates high negative attention
        (preventing information flow), 0.0 indicates no mask effect.
    """  # noqa: E501

    img_mask = np.zeros((1, depth, height, width, 1))
    cnt = 0
    for d in (
        slice(-window_size[0]),
        slice(-window_size[0], -shift_size[0]),
        slice(-shift_size[0], None),
    ):
        for h in (
            slice(-window_size[1]),
            slice(-window_size[1], -shift_size[1]),
            slice(-shift_size[1], None),
        ):
            for w in (
                slice(-window_size[2]),
                slice(-window_size[2], -shift_size[2]),
                slice(-shift_size[2], None),
            ):
                img_mask[:, d, h, w, :] = cnt
                cnt = cnt + 1
    mask_windows = window_partition(img_mask, window_size)
    mask_windows = ops.squeeze(mask_windows, axis=-1)
    attn_mask = ops.expand_dims(mask_windows, axis=1) - ops.expand_dims(
        mask_windows, axis=2
    )
    attn_mask = ops.where(attn_mask != 0, -100.0, attn_mask)
    attn_mask = ops.where(attn_mask == 0, 0.0, attn_mask)
    return attn_mask


class MLP(keras.layers.Layer):
    """A Multilayer perceptron(MLP) layer.

    Args:
        hidden_dim (int): The number of units in the hidden layer.
        output_dim (int): The number of units in the output layer.
        drop_rate  (float): Float between 0 and 1. Fraction of the
            input units to drop.
        activation (str): Activation to use in the hidden layers.
            Default is `"gelu"`.

    References:
        - [Video Swin Transformer](https://arxiv.org/abs/2106.13230)
        - [Video Swin Transformer GitHub](https://github.com/SwinTransformer/Video-Swin-Transformer)
    """  # noqa: E501

    def __init__(
        self, hidden_dim, output_dim, drop_rate=0.0, activation="gelu", **kwargs
    ):
        super().__init__(**kwargs)
        self.output_dim = output_dim
        self.hidden_dim = hidden_dim
        self._activation_identifier = activation
        self.drop_rate = drop_rate
        self.activation = keras.layers.Activation(self._activation_identifier)
        self.fc1 = keras.layers.Dense(self.hidden_dim)
        self.fc2 = keras.layers.Dense(self.output_dim)
        self.dropout = keras.layers.Dropout(self.drop_rate)

    def build(self, input_shape):
        self.fc1.build(input_shape)
        self.fc2.build((*input_shape[:-1], self.hidden_dim))
        self.built = True

    def call(self, x, training=None):
        x = self.fc1(x)
        x = self.activation(x)
        x = self.dropout(x, training=training)
        x = self.fc2(x)
        x = self.dropout(x, training=training)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "output_dim": self.output_dim,
                "hidden_dim": self.hidden_dim,
                "drop_rate": self.drop_rate,
                "activation": self._activation_identifier,
            }
        )
        return config


class VideoSwinPatchingAndEmbedding(keras.Model):
    """Video to Patch Embedding layer for Video Swin Transformer models.

    This layer performs the initial step in a Video Swin Transformer architecture by
    partitioning the input video into 3D patches and embedding them into a vector
    dimensional space.

    Args:
        patch_size (int): Size of the patch along each dimension
            (depth, height, width). Default: (2,4,4).
        embed_dim (int): Number of linear projection output channels. Default: 96.
        norm_layer (keras.layers, optional): Normalization layer. Default: None

    References:
        - [Video Swin Transformer](https://arxiv.org/abs/2106.13230)
        - [Video Swin Transformer GitHub](https://github.com/SwinTransformer/Video-Swin-Transformer)
    """  # noqa: E501

    def __init__(
        self, patch_size=(2, 4, 4), embed_dim=96, norm_layer=None, **kwargs
    ):
        super().__init__(**kwargs)
        self.patch_size = patch_size
        self.embed_dim = embed_dim
        self.norm_layer = norm_layer

    def __compute_padding(self, dim, patch_size):
        pad_amount = patch_size - (dim % patch_size)
        return [0, pad_amount if pad_amount != patch_size else 0]

    def build(self, input_shape):
        self.pads = [
            [0, 0],
            self.__compute_padding(input_shape[1], self.patch_size[0]),
            self.__compute_padding(input_shape[2], self.patch_size[1]),
            self.__compute_padding(input_shape[3], self.patch_size[2]),
            [0, 0],
        ]

        if self.norm_layer is not None:
            self.norm = self.norm_layer(
                axis=-1, epsilon=1e-5, name="embed_norm"
            )
            self.norm.build((None, None, None, None, self.embed_dim))

        self.proj = keras.layers.Conv3D(
            self.embed_dim,
            kernel_size=self.patch_size,
            strides=self.patch_size,
            name="embed_proj",
        )
        self.proj.build((None, None, None, None, input_shape[-1]))
        self.built = True

    def call(self, x):
        x = ops.pad(x, self.pads)
        x = self.proj(x)

        if self.norm_layer is not None:
            x = self.norm(x)

        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "patch_size": self.patch_size,
                "embed_dim": self.embed_dim,
            }
        )
        return config


class VideoSwinPatchMerging(keras.layers.Layer):
    """Patch Merging Layer in Video Swin Transformer models.

    This layer performs a downsampling step by merging four neighboring patches
    from the previous layer into a single patch in the output. It achieves this
    by concatenation and linear projection.

    Args:
        input_dim (int): Number of input channels in the feature maps.
        norm_layer (keras.layers, optional): Normalization layer.
            Default: LayerNormalization

    References:
        - [Video Swin Transformer](https://arxiv.org/abs/2106.13230)
        - [Video Swin Transformer GitHub](https://github.com/SwinTransformer/Video-Swin-Transformer)
    """  # noqa: E501

    def __init__(self, input_dim, norm_layer=None, **kwargs):
        super().__init__(**kwargs)
        self.input_dim = input_dim
        self.norm_layer = norm_layer

    def build(self, input_shape):
        batch_size, depth, height, width, channel = input_shape
        self.reduction = keras.layers.Dense(2 * self.input_dim, use_bias=False)
        self.reduction.build(
            (batch_size, depth, height // 2, width // 2, 4 * channel)
        )

        if self.norm_layer is not None:
            self.norm = self.norm_layer(axis=-1, epsilon=1e-5)
            self.norm.build(
                (batch_size, depth, height // 2, width // 2, 4 * channel)
            )

        # compute padding if needed
        self.pads = [
            [0, 0],
            [0, 0],
            [0, ops.mod(height, 2)],
            [0, ops.mod(width, 2)],
            [0, 0],
        ]
        self.built = True

    def call(self, x):
        # padding if needed
        x = ops.pad(x, self.pads)
        x0 = x[:, :, 0::2, 0::2, :]  # B D H/2 W/2 C
        x1 = x[:, :, 1::2, 0::2, :]  # B D H/2 W/2 C
        x2 = x[:, :, 0::2, 1::2, :]  # B D H/2 W/2 C
        x3 = x[:, :, 1::2, 1::2, :]  # B D H/2 W/2 C
        x = ops.concatenate([x0, x1, x2, x3], axis=-1)  # B D H/2 W/2 4*C

        if self.norm_layer is not None:
            x = self.norm(x)

        x = self.reduction(x)
        return x

    def compute_output_shape(self, input_shape):
        batch_size, depth, height, width, _ = input_shape
        return (batch_size, depth, height // 2, width // 2, 2 * self.input_dim)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "input_dim": self.input_dim,
            }
        )
        return config


class VideoSwinWindowAttention(keras.Model):
    """It tackles long-range video dependencies by splitting features into windows
    and using relative position bias within each window for focused attention.
    It supports both of shifted and non-shifted window.

    Args:
        input_dim (int): The number of input channels in the feature maps.
        window_size (tuple[int]): The temporal length, height and width of the window.
        num_heads (int): Number of attention heads.
        qkv_bias (bool, optional):  If True, add a learnable bias to query, key, value. Default: True
        qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set
        attn_drop_rate (float, optional): Dropout ratio of attention weight. Default: 0.0
        proj_drop_rate (float, optional): Dropout ratio of output. Default: 0.0

    References:
        - [Video Swin Transformer](https://arxiv.org/abs/2106.13230)
        - [Video Swin Transformer GitHub](https://github.com/SwinTransformer/Video-Swin-Transformer)
    """  # noqa: E501

    def __init__(
        self,
        input_dim,
        window_size,
        num_heads,
        qkv_bias=True,
        qk_scale=None,
        attn_drop_rate=0.0,
        proj_drop_rate=0.0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        # variables
        self.input_dim = input_dim
        self.window_size = window_size
        self.num_heads = num_heads
        head_dim = input_dim // num_heads
        self.qk_scale = qk_scale
        self.scale = qk_scale or head_dim**-0.5
        self.qkv_bias = qkv_bias
        self.attn_drop_rate = attn_drop_rate
        self.proj_drop_rate = proj_drop_rate

    def get_relative_position_index(
        self, window_depth, window_height, window_width
    ):
        y_y, z_z, x_x = ops.meshgrid(
            ops.arange(window_width),
            ops.arange(window_depth),
            ops.arange(window_height),
        )
        coords = ops.stack([z_z, y_y, x_x], axis=0)
        coords_flatten = ops.reshape(coords, [3, -1])
        relative_coords = (
            coords_flatten[:, :, None] - coords_flatten[:, None, :]
        )
        relative_coords = ops.transpose(relative_coords, axes=[1, 2, 0])
        z_z = (
            (relative_coords[:, :, 0] + window_depth - 1)
            * (2 * window_height - 1)
            * (2 * window_width - 1)
        )
        x_x = (relative_coords[:, :, 1] + window_height - 1) * (
            2 * window_width - 1
        )
        y_y = relative_coords[:, :, 2] + window_width - 1
        relative_coords = ops.stack([z_z, x_x, y_y], axis=-1)
        return ops.sum(relative_coords, axis=-1)

    def build(self, input_shape):
        self.relative_position_bias_table = self.add_weight(
            shape=(
                (2 * self.window_size[0] - 1)
                * (2 * self.window_size[1] - 1)
                * (2 * self.window_size[2] - 1),
                self.num_heads,
            ),
            initializer="zeros",
            trainable=True,
            name="relative_position_bias_table",
        )
        self.relative_position_index = self.get_relative_position_index(
            self.window_size[0], self.window_size[1], self.window_size[2]
        )

        # layers
        self.qkv = keras.layers.Dense(
            self.input_dim * 3, use_bias=self.qkv_bias
        )
        self.attn_drop = keras.layers.Dropout(self.attn_drop_rate)
        self.proj = keras.layers.Dense(self.input_dim)
        self.proj_drop = keras.layers.Dropout(self.proj_drop_rate)
        self.qkv.build(input_shape)
        self.proj.build(input_shape)
        self.built = True

    def call(self, x, mask=None, training=None):
        input_shape = ops.shape(x)
        batch_size, depth, channel = (
            input_shape[0],
            input_shape[1],
            input_shape[2],
        )

        qkv = self.qkv(x)
        qkv = ops.reshape(
            qkv,
            [batch_size, depth, 3, self.num_heads, channel // self.num_heads],
        )
        qkv = ops.transpose(qkv, [2, 0, 3, 1, 4])
        q, k, v = ops.split(qkv, 3, axis=0)
        q = ops.squeeze(q, axis=0) * self.scale
        k = ops.squeeze(k, axis=0)
        v = ops.squeeze(v, axis=0)
        attn = ops.matmul(q, ops.transpose(k, [0, 1, 3, 2]))

        rel_pos_bias = ops.take(
            self.relative_position_bias_table,
            self.relative_position_index[:depth, :depth],
            axis=0,
        )
        rel_pos_bias = ops.reshape(rel_pos_bias, [depth, depth, -1])
        rel_pos_bias = ops.transpose(rel_pos_bias, [2, 0, 1])
        attn = attn + rel_pos_bias[None, ...]

        if mask is not None:
            mask_size = ops.shape(mask)[0]
            mask = ops.cast(mask, dtype=attn.dtype)
            attn = (
                ops.reshape(
                    attn,
                    [
                        batch_size // mask_size,
                        mask_size,
                        self.num_heads,
                        depth,
                        depth,
                    ],
                )
                + mask[:, None, :, :]
            )
            attn = ops.reshape(attn, [-1, self.num_heads, depth, depth])

        attn = keras.activations.softmax(attn, axis=-1)
        attn = self.attn_drop(attn, training=training)
        x = ops.matmul(attn, v)
        x = ops.transpose(x, [0, 2, 1, 3])
        x = ops.reshape(x, [batch_size, depth, channel])
        x = self.proj(x)
        x = self.proj_drop(x, training=training)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "input_dim": self.input_dim,
                "window_size": self.window_size,
                "num_heads": self.num_heads,
                "qk_scale": self.qk_scale,
                "qkv_bias": self.qkv_bias,
                "attn_drop_rate": self.attn_drop_rate,
                "proj_drop_rate": self.proj_drop_rate,
            }
        )
        return config


class VideoSwinBasicLayer(keras.Model):
    """A basic Video Swin Transformer layer for one stage.

    Args:
        input_dim (int): Number of feature channels
        depth (int): Depths of this stage.
        num_heads (int): Number of attention head.
        window_size (tuple[int]): Local window size. Default: (1,7,7).
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim. Default: 4.
        qkv_bias (bool, optional): If True, add a learnable bias to query, key, value. Default: True
        qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set.
        drop (float, optional): Dropout rate. Default: 0.0
        attn_drop (float, optional): Attention dropout rate. Default: 0.0
        drop_path (float | tuple[float], optional): Stochastic depth rate. Default: 0.0
        norm_layer (keras.layers, optional): Normalization layer. Default: LayerNormalization
        downsample (keras.layers | None, optional): Downsample layer at the end of the layer. Default: None

    References:
        - [Video Swin Transformer](https://arxiv.org/abs/2106.13230)
        - [Video Swin Transformer GitHub](https://github.com/SwinTransformer/Video-Swin-Transformer)
    """  # noqa: E501

    def __init__(
        self,
        input_dim,
        depth,
        num_heads,
        window_size=(1, 7, 7),
        mlp_ratio=4.0,
        qkv_bias=False,
        qk_scale=None,
        drop_rate=0.0,
        attn_drop_rate=0.0,
        drop_path_rate=0.0,
        norm_layer=None,
        downsampling_layer=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.input_dim = input_dim
        self.num_heads = num_heads
        self.window_size = window_size
        self.mlp_ratio = mlp_ratio
        self.shift_size = tuple([i // 2 for i in window_size])
        self.depth = depth
        self.qkv_bias = qkv_bias
        self.qk_scale = qk_scale
        self.drop_rate = drop_rate
        self.attn_drop_rate = attn_drop_rate
        self.drop_path_rate = drop_path_rate
        self.norm_layer = norm_layer
        self.downsampling_layer = downsampling_layer

    def __compute_dim_padded(self, input_dim, window_dim_size):
        input_dim = ops.cast(input_dim, dtype="float32")
        window_dim_size = ops.cast(window_dim_size, dtype="float32")
        return ops.cast(
            ops.ceil(input_dim / window_dim_size) * window_dim_size, "int32"
        )

    def build(self, input_shape):
        self.window_size, self.shift_size = get_window_size(
            input_shape[1:-1], self.window_size, self.shift_size
        )
        self.depth_pad = self.__compute_dim_padded(
            input_shape[1], self.window_size[0]
        )
        self.height_pad = self.__compute_dim_padded(
            input_shape[2], self.window_size[1]
        )
        self.width_pad = self.__compute_dim_padded(
            input_shape[3], self.window_size[2]
        )
        self.attn_mask = compute_mask(
            self.depth_pad,
            self.height_pad,
            self.width_pad,
            self.window_size,
            self.shift_size,
        )

        # build blocks
        self.blocks = [
            VideoSwinTransformerBlock(
                self.input_dim,
                num_heads=self.num_heads,
                window_size=self.window_size,
                shift_size=(0, 0, 0) if (i % 2 == 0) else self.shift_size,
                mlp_ratio=self.mlp_ratio,
                qkv_bias=self.qkv_bias,
                qk_scale=self.qk_scale,
                drop_rate=self.drop_rate,
                attn_drop_rate=self.attn_drop_rate,
                drop_path_rate=(
                    self.drop_path_rate[i]
                    if isinstance(self.drop_path_rate, list)
                    else self.drop_path_rate
                ),
                norm_layer=self.norm_layer,
            )
            for i in range(self.depth)
        ]

        if self.downsampling_layer is not None:
            self.downsample = self.downsampling_layer(
                input_dim=self.input_dim, norm_layer=self.norm_layer
            )
            self.downsample.build(input_shape)

        for i in range(self.depth):
            self.blocks[i].build(input_shape)

        self.built = True

    def compute_output_shape(self, input_shape):
        if self.downsampling_layer is not None:
            input_shape = self.downsample.compute_output_shape(input_shape)
            return input_shape

        return input_shape

    def call(self, x, training=None):
        input_shape = ops.shape(x)
        batch_size, depth, height, width, channel = (
            input_shape[0],
            input_shape[1],
            input_shape[2],
            input_shape[3],
            input_shape[4],
        )

        for block in self.blocks:
            x = block(x, self.attn_mask, training=training)

        x = ops.reshape(x, [batch_size, depth, height, width, channel])

        if self.downsampling_layer is not None:
            x = self.downsample(x)

        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "input_dim": self.input_dim,
                "window_size": self.window_size,
                "num_heads": self.num_heads,
                "mlp_ratio": self.mlp_ratio,
                "depth": self.depth,
                "qkv_bias": self.qkv_bias,
                "qk_scale": self.qk_scale,
                "drop_rate": self.drop_rate,
                "attn_drop_rate": self.attn_drop_rate,
                "drop_path_rate": self.drop_path_rate,
            }
        )
        return config


class VideoSwinTransformerBlock(keras.Model):
    """Video Swin Transformer Block.

    Args:
        input_dim (int): Number of feature channels.
        num_heads (int): Number of attention heads.
        window_size (tuple[int]): Local window size. Default: (2, 7, 7)
        shift_size (tuple[int]): Shift size for SW-MSA. Default: (0, 0, 0)
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
            Default: 4.0
        qkv_bias (bool, optional): If True, add a learnable bias to query, key, value.
            Default: True
        qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set.
            Default: None
        drop (float, optional): Dropout rate. Default: 0.0
        attn_drop (float, optionalc): Attention dropout rate. Default: 0.0
        drop_path (float, optional): Stochastic depth rate. Default: 0.0
        act_layer (keras.layers.Activation, optional): Activation layer. Default: gelu
        norm_layer (keras.layers, optional): Normalization layer.
            Default: LayerNormalization

    References:
        - [Video Swin Transformer](https://arxiv.org/abs/2106.13230)
        - [Video Swin Transformer GitHub](https://github.com/SwinTransformer/Video-Swin-Transformer)
    """  # noqa: E501

    def __init__(
        self,
        input_dim,
        num_heads,
        window_size=(2, 7, 7),
        shift_size=(0, 0, 0),
        mlp_ratio=4.0,
        qkv_bias=True,
        qk_scale=None,
        drop_rate=0.0,
        attn_drop_rate=0.0,
        drop_path_rate=0.0,
        activation="gelu",
        norm_layer=keras.layers.LayerNormalization,
        **kwargs,
    ):
        super().__init__(**kwargs)
        # variables
        self.input_dim = input_dim
        self.num_heads = num_heads
        self.window_size = window_size
        self.shift_size = shift_size
        self.mlp_ratio = mlp_ratio
        self.qkv_bias = qkv_bias
        self.qk_scale = qk_scale
        self.drop_rate = drop_rate
        self.attn_drop_rate = attn_drop_rate
        self.drop_path_rate = drop_path_rate
        self.mlp_hidden_dim = int(input_dim * mlp_ratio)
        self.norm_layer = norm_layer
        self._activation_identifier = activation

        for i, (shift, window) in enumerate(
            zip(self.shift_size, self.window_size)
        ):
            if not (0 <= shift < window):
                raise ValueError(
                    f"shift_size[{i}] must be in the range 0 to less than "
                    f"window_size[{i}], but got shift_size[{i}]={shift} "
                    f"and window_size[{i}]={window}."
                )

    def build(self, input_shape):
        self.window_size, self.shift_size = get_window_size(
            input_shape[1:-1], self.window_size, self.shift_size
        )
        self.apply_cyclic_shift = any(i > 0 for i in self.shift_size)

        # layers
        self.drop_path = (
            DropPath(self.drop_path_rate)
            if self.drop_path_rate > 0.0
            else keras.layers.Identity()
        )

        self.norm1 = self.norm_layer(axis=-1, epsilon=1e-05)
        self.norm1.build(input_shape)

        self.attn = VideoSwinWindowAttention(
            self.input_dim,
            window_size=self.window_size,
            num_heads=self.num_heads,
            qkv_bias=self.qkv_bias,
            qk_scale=self.qk_scale,
            attn_drop_rate=self.attn_drop_rate,
            proj_drop_rate=self.drop_rate,
        )
        self.attn.build((None, None, self.input_dim))

        self.norm2 = self.norm_layer(axis=-1, epsilon=1e-05)
        self.norm2.build((*input_shape[:-1], self.input_dim))

        self.mlp = MLP(
            output_dim=self.input_dim,
            hidden_dim=self.mlp_hidden_dim,
            activation=self._activation_identifier,
            drop_rate=self.drop_rate,
        )
        self.mlp.build((*input_shape[:-1], self.input_dim))

        # compute padding if needed.
        # pad input feature maps to multiples of window size.
        _, depth, height, width, _ = input_shape
        pad_l = pad_t = pad_d0 = 0
        self.pad_d1 = ops.mod(-depth + self.window_size[0], self.window_size[0])
        self.pad_b = ops.mod(-height + self.window_size[1], self.window_size[1])
        self.pad_r = ops.mod(-width + self.window_size[2], self.window_size[2])
        self.pads = [
            [0, 0],
            [pad_d0, self.pad_d1],
            [pad_t, self.pad_b],
            [pad_l, self.pad_r],
            [0, 0],
        ]
        self.apply_pad = any(
            value > 0 for value in (self.pad_d1, self.pad_r, self.pad_b)
        )
        self.built = True

    def first_forward(self, x, mask_matrix, training):
        input_shape = ops.shape(x)
        batch_size, depth, height, width, _ = (
            input_shape[0],
            input_shape[1],
            input_shape[2],
            input_shape[3],
            input_shape[4],
        )
        x = self.norm1(x)

        # apply padding if needed.
        x = ops.pad(x, self.pads)

        input_shape = ops.shape(x)
        depth_pad, height_pad, width_pad = (
            input_shape[1],
            input_shape[2],
            input_shape[3],
        )

        # cyclic shift
        if self.apply_cyclic_shift:
            shifted_x = ops.roll(
                x,
                shift=(
                    -self.shift_size[0],
                    -self.shift_size[1],
                    -self.shift_size[2],
                ),
                axis=(1, 2, 3),
            )
            attn_mask = mask_matrix
        else:
            shifted_x = x
            attn_mask = None

        # partition windows
        x_windows = window_partition(shifted_x, self.window_size)

        # get attentions params
        attn_windows = self.attn(x_windows, mask=attn_mask, training=training)

        # reverse the swin windows
        shifted_x = window_reverse(
            attn_windows,
            self.window_size,
            batch_size,
            depth_pad,
            height_pad,
            width_pad,
        )

        # reverse cyclic shift
        if self.apply_cyclic_shift:
            x = ops.roll(
                shifted_x,
                shift=(
                    self.shift_size[0],
                    self.shift_size[1],
                    self.shift_size[2],
                ),
                axis=(1, 2, 3),
            )
        else:
            x = shifted_x

        # pad if required
        if self.apply_pad:
            return x[:, :depth, :height, :width, :]

        return x

    def second_forward(self, x, training):
        x = self.norm2(x)
        x = self.mlp(x)
        x = self.drop_path(x, training=training)
        return x

    def call(self, x, mask_matrix=None, training=None):
        shortcut = x
        x = self.first_forward(x, mask_matrix, training)
        x = shortcut + self.drop_path(x)
        x = x + self.second_forward(x, training)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "input_dim": self.input_dim,
                "window_size": self.num_heads,
                "num_heads": self.window_size,
                "shift_size": self.shift_size,
                "mlp_ratio": self.mlp_ratio,
                "qkv_bias": self.qkv_bias,
                "qk_scale": self.qk_scale,
                "drop_rate": self.drop_rate,
                "attn_drop_rate": self.attn_drop_rate,
                "drop_path_rate": self.drop_path_rate,
                "mlp_hidden_dim": self.mlp_hidden_dim,
                "activation": self._activation_identifier,
            }
        )
        return config



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_cv.src.keypoint.converters import convert_format
from keras_cv.src.keypoint.formats import REL_XY
from keras_cv.src.keypoint.formats import XY
from keras_cv.src.keypoint.utils import filter_out_of_image



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.layers.preprocessing_3d.base_augmentation_layer_3d import (
    BaseAugmentationLayer3D,
)
from keras_cv.src.layers.preprocessing_3d.waymo.frustum_random_dropping_points import (  # noqa: E501
    FrustumRandomDroppingPoints,
)
from keras_cv.src.layers.preprocessing_3d.waymo.frustum_random_point_feature_noise import (  # noqa: E501
    FrustumRandomPointFeatureNoise,
)
from keras_cv.src.layers.preprocessing_3d.waymo.global_random_dropping_points import (  # noqa: E501
    GlobalRandomDroppingPoints,
)
from keras_cv.src.layers.preprocessing_3d.waymo.global_random_flip import (
    GlobalRandomFlip,
)
from keras_cv.src.layers.preprocessing_3d.waymo.global_random_rotation import (
    GlobalRandomRotation,
)
from keras_cv.src.layers.preprocessing_3d.waymo.global_random_scaling import (
    GlobalRandomScaling,
)
from keras_cv.src.layers.preprocessing_3d.waymo.global_random_translation import (  # noqa: E501
    GlobalRandomTranslation,
)
from keras_cv.src.layers.preprocessing_3d.waymo.group_points_by_bounding_boxes import (  # noqa: E501
    GroupPointsByBoundingBoxes,
)
from keras_cv.src.layers.preprocessing_3d.waymo.random_copy_paste import (
    RandomCopyPaste,
)
from keras_cv.src.layers.preprocessing_3d.waymo.random_drop_box import (
    RandomDropBox,
)
from keras_cv.src.layers.preprocessing_3d.waymo.swap_background import (
    SwapBackground,
)



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.layers.preprocessing.base_image_augmentation_layer import (
    BaseImageAugmentationLayer,
)


@keras_cv_export("keras_cv.layers.RepeatedAugmentation")
class RepeatedAugmentation(BaseImageAugmentationLayer):
    """RepeatedAugmentation augments each image in a batch multiple times.

    This technique exists to emulate the behavior of stochastic gradient descent
    within the context of mini-batch gradient descent. When training large
    vision models, choosing a large batch size can introduce too much noise into
    aggregated gradients causing the overall batch's gradients to be less
    effective than gradients produced using smaller gradients.
    RepeatedAugmentation handles this by re-using the same image multiple times
    within a batch creating correlated samples.

    This layer increases your batch size by a factor of `len(augmenters)`.

    Args:
        augmenters: the augmenters to use to augment the image
        shuffle: whether to shuffle the result. Essential when using an
            asynchronous distribution strategy such as ParameterServerStrategy.

    Example:

    List of identical augmenters:
    ```python
    repeated_augment = cv_layers.RepeatedAugmentation(
        augmenters=[cv_layers.RandAugment(value_range=(0, 255))] * 8
    )
    inputs = {
        "images": tf.ones((8, 512, 512, 3)),
        "labels": tf.ones((8,)),
    }
    outputs = repeated_augment(inputs)
    # outputs now has a batch size of 64 because there are 8 augmenters
    ```

    List of distinct augmenters:
    ```python
    repeated_augment = cv_layers.RepeatedAugmentation(
        augmenters=[
            cv_layers.RandAugment(value_range=(0, 255)),
            cv_layers.RandomFlip(),
        ]
    )
    inputs = {
        "images": tf.ones((8, 512, 512, 3)),
        "labels": tf.ones((8,)),
    }
    outputs = repeated_augment(inputs)
    ```

    References:
    - [DEIT implementation](https://github.com/facebookresearch/deit/blob/ee8893c8063f6937fec7096e47ba324c206e22b9/samplers.py#L8)
    - [Original publication](https://openaccess.thecvf.com/content_CVPR_2020/papers/Hoffer_Augment_Your_Batch_Improving_Generalization_Through_Instance_Repetition_CVPR_2020_paper.pdf)

    """  # noqa: E501

    def __init__(self, augmenters, shuffle=True, **kwargs):
        super().__init__(**kwargs)
        self.augmenters = augmenters
        self.shuffle = shuffle

    def _batch_augment(self, inputs):
        if "bounding_boxes" in inputs:
            raise ValueError(
                "RepeatedAugmentation() does not yet support bounding box "
                "labels."
            )

        augmenter_outputs = [augmenter(inputs) for augmenter in self.augmenters]

        outputs = {}
        for k in inputs.keys():
            outputs[k] = tf.concat(
                [output[k] for output in augmenter_outputs], axis=0
            )

        if not self.shuffle:
            return outputs
        return self.shuffle_outputs(outputs)

    def shuffle_outputs(self, result):
        indices = tf.range(
            start=0, limit=tf.shape(result["images"])[0], dtype=tf.int32
        )
        indices = tf.random.shuffle(indices)
        for key in result:
            result[key] = tf.gather(result[key], indices)
        return result

    def _augment(self, inputs):
        raise ValueError(
            "RepeatedAugmentation() only works in batched mode. If "
            "you would like to create batches from a single image, use "
            "`x = tf.expand_dims(x, axis=0)` on your input images and labels."
        )

    def get_config(self):
        config = super().get_config()
        config.update({"augmenters": self.augmenters, "shuffle": self.shuffle})
        return config

    @classmethod
    def from_config(cls, config):
        if config["augmenters"] and isinstance(config["augmenters"][0], dict):
            config["augmenters"] = keras.utils.deserialize_keras_object(
                config["augmenters"]
            )
        return cls(**config)



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from keras_cv.src.backend import ops


def _decode_strings_to_utf8(inputs):
    """Recursively decodes to list of strings with 'utf-8' encoding."""
    if isinstance(inputs, bytes):
        # Handles the case when the input is a scalar string.
        return inputs.decode("utf-8", errors="ignore")
    else:
        # Recursively iterate when input is a list.
        return [_decode_strings_to_utf8(x) for x in inputs]


def tensor_to_list(inputs):
    """Converts a tensor to nested lists.

    Args:
        inputs: Input tensor, or dict/list/tuple of input tensors.
    """
    if not isinstance(inputs, (tf.RaggedTensor, tf.Tensor)):
        inputs = tf.convert_to_tensor(inputs)
    if isinstance(inputs, tf.RaggedTensor):
        list_outputs = inputs.to_list()
    elif isinstance(inputs, tf.Tensor):
        list_outputs = inputs.numpy()
        if inputs.shape.rank != 0:
            list_outputs = list_outputs.tolist()
    if inputs.dtype == tf.string:
        list_outputs = _decode_strings_to_utf8(list_outputs)
    return list_outputs


def convert_to_backend_tensor_or_python_list(x):
    """
    Convert a tensor to the backend friendly representation of the data.

    This wraps `ops.convert_to_tensor` to account for the fact that torch and
    jax both lack native types for ragged and string data.

    If we encounter one of these types in torch or jax, we will instead covert
    the tensor to simple pythonic types (lists of strings).
    """
    if isinstance(x, tf.RaggedTensor) or getattr(x, "dtype", None) == tf.string:
        return tensor_to_list(x)
    return ops.convert_to_tensor(x)


def convert_inputs_to_list_of_tensor_segments(x):
    """Converts user inputs to a list of a tensor segments.

    For models and layers which accept lists of string tensors to pack together,
    this method converts user inputs to a uniform format in a way that can be
    considered canonical for the library.

    We handle the following:

    - A single string will be converted to a tensor and wrapped in a list.
    - A list of strings will be converted to a tensor and wrapped in a list.
    - A single tensor will be wrapped in a list.
    - A list of tensors will be passed through unaltered.

    All other inputs will result in an error. This effectively means that users
    who would like to pack multiple segments together should convert those
    segments to tensors before calling the layer. This removes any ambiguity
    in the input for those cases.
    """
    # Check the input type.
    is_string = isinstance(x, (str, bytes))
    is_tensor = hasattr(x, "__array__")
    is_string_list = (
        isinstance(x, (list, tuple)) and x and isinstance(x[0], (str, bytes))
    )
    is_tensor_list = (
        isinstance(x, (list, tuple)) and x and hasattr(x[0], "__array__")
    )

    if is_string or is_string_list:
        # Automatically convert raw strings or string lists to tensors.
        # Wrap this input as a single (possibly batched) segment.
        x = [tf.convert_to_tensor(x)]
    elif is_tensor:
        # Automatically wrap a single tensor as a single segment.
        x = [x]
    elif is_tensor_list:
        # Pass lists of tensors though unaltered.
        x = x
    else:
        # Error for all other input.
        raise ValueError(
            f"Unsupported input for `x`. `x` should be a string, a list of "
            "strings, or a list of tensors. If passing multiple segments "
            "which should packed together, please convert your inputs to a "
            f"list of tensors. Received `x={x}`"
        )
    return x



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.models.segmentation.segformer.segformer import SegFormer
from keras_cv.src.models.segmentation.segformer.segformer_presets import presets
from keras_cv.src.utils.python_utils import classproperty

ALIAS_DOCSTRING = """SegFormer model.

    For transfer learning use cases, make sure to read the
    [guide to transfer learning & fine-tuning](https://keras.io/guides/transfer_learning/).

    Args:
        backbone: a KerasCV backbone for feature extraction.
        num_classes: the number of classes for segmentation, including the background class.

    Example:
    ```python
    input_data = tf.ones(shape=(8, 224, 224, 3))

    # Randomly initialized backbone
    backbone = keras_cv.models.MiTBackbone.from_preset("mit_b0_imagenet")
    segformer = keras_cv.models.SegFormer(backbone=backbone, num_classes=19)
    output = model(input_data)
    ```
"""  # noqa: E501


@keras_cv_export("keras_cv.models.SegFormerB0")
class SegFormerB0(SegFormer):
    def __new__(
        cls,
        num_classes,
        **kwargs,
    ):
        # Pack args in kwargs
        kwargs.update(
            {
                "num_classes": num_classes,
            }
        )
        return SegFormer.from_preset("segformer_b0", **kwargs)

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return {
            "segformer_b0": copy.deepcopy(presets["segformer_b0"]),
        }

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return cls.presets


@keras_cv_export("keras_cv.models.SegFormerB1")
class SegFormerB1(SegFormer):
    def __new__(
        cls,
        num_classes,
        **kwargs,
    ):
        # Pack args in kwargs
        kwargs.update(
            {
                "num_classes": num_classes,
            }
        )
        return SegFormer.from_preset("segformer_b1", **kwargs)

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return {
            "segformer_b1": copy.deepcopy(presets["segformer_b1"]),
        }

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return cls.presets


@keras_cv_export("keras_cv.models.SegFormerB2")
class SegFormerB2(SegFormer):
    def __new__(
        cls,
        num_classes,
        **kwargs,
    ):
        # Pack args in kwargs
        kwargs.update(
            {
                "num_classes": num_classes,
            }
        )
        return SegFormer.from_preset("segformer_b2", **kwargs)

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return {
            "segformer_b2": copy.deepcopy(presets["segformer_b2"]),
        }

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return cls.presets


@keras_cv_export("keras_cv.models.SegFormerB3")
class SegFormerB3(SegFormer):
    def __new__(
        cls,
        num_classes,
        **kwargs,
    ):
        # Pack args in kwargs
        kwargs.update(
            {
                "num_classes": num_classes,
            }
        )
        return SegFormer.from_preset("segformer_b3", **kwargs)

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return {
            "segformer_b3": copy.deepcopy(presets["segformer_b3"]),
        }

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return cls.presets


@keras_cv_export("keras_cv.models.SegFormerB4")
class SegFormerB4(SegFormer):
    def __new__(
        cls,
        num_classes,
        **kwargs,
    ):
        # Pack args in kwargs
        kwargs.update(
            {
                "num_classes": num_classes,
            }
        )
        return SegFormer.from_preset("segformer_b4", **kwargs)

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return {
            "segformer_b4": copy.deepcopy(presets["segformer_b4"]),
        }

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return cls.presets


@keras_cv_export("keras_cv.models.SegFormerB5")
class SegFormerB5(SegFormer):
    def __new__(
        cls,
        num_classes,
        **kwargs,
    ):
        # Pack args in kwargs
        kwargs.update(
            {
                "num_classes": num_classes,
            }
        )
        return SegFormer.from_preset("segformer_b5", **kwargs)

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return {
            "segformer_b5": copy.deepcopy(presets["segformer_b5"]),
        }

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return cls.presets


setattr(
    SegFormerB0,
    "__doc__",
    ALIAS_DOCSTRING.format(name="SegFormerB0"),
)

setattr(
    SegFormerB1,
    "__doc__",
    ALIAS_DOCSTRING.format(name="SegFormerB1"),
)

setattr(
    SegFormerB2,
    "__doc__",
    ALIAS_DOCSTRING.format(name="SegFormerB2"),
)

setattr(
    SegFormerB3,
    "__doc__",
    ALIAS_DOCSTRING.format(name="SegFormerB3"),
)

setattr(
    SegFormerB4,
    "__doc__",
    ALIAS_DOCSTRING.format(name="SegFormerB4"),
)

setattr(
    SegFormerB5,
    "__doc__",
    ALIAS_DOCSTRING.format(name="SegFormerB5"),
)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MLP Mixer models for KerasCV.

Reference:
  - [MLP-Mixer: An all-MLP Architecture for Vision](https://arxiv.org/abs/2105.01601)
"""  # noqa: E501

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import backend
from tensorflow.keras import layers

from keras_cv.src.models.legacy import utils

MODEL_CONFIGS = {
    "MLPMixerB16": {
        "patch_size": 16,
        "num_blocks": 12,
        "hidden_dim": 768,
        "tokens_mlp_dim": 384,
        "channels_mlp_dim": 3072,
    },
    "MLPMixerB32": {
        "patch_size": 32,
        "num_blocks": 12,
        "hidden_dim": 768,
        "tokens_mlp_dim": 384,
        "channels_mlp_dim": 3072,
    },
    "MLPMixerL16": {
        "patch_size": 16,
        "num_blocks": 24,
        "hidden_dim": 1024,
        "tokens_mlp_dim": 512,
        "channels_mlp_dim": 4096,
    },
}

BASE_DOCSTRING = """Instantiates the {name} architecture.

    Reference:
        - [MLP-Mixer: An all-MLP Architecture for Vision](https://arxiv.org/abs/2105.01601)

    This class represents a Keras {name} model.

    For transfer learning use cases, make sure to read the [guide to transfer
        learning & fine-tuning](https://keras.io/guides/transfer_learning/).

    Args:
        include_rescaling: bool, whether to rescale the inputs. If set to
            True, inputs will be passed through a `Rescaling(1/255.0)` layer.
        include_top: bool, whether to include the fully-connected layer at the
            top of the network. If provided, num_classes must be provided.
        num_classes: integer, optional number of classes to classify images
            into. Only to be specified if `include_top` is True.
        weights: one of `None` (random initialization), a pretrained weight file
            path, or a reference to pre-trained weights (e.g.
            'imagenet/classification')(see available pre-trained weights in
            weights.py)
        input_shape: optional shape tuple, defaults to (None, None, 3).
        input_tensor: optional Keras tensor (i.e., output of `layers.Input()`)
            to use as image input for the model.
        pooling: optional pooling mode for feature extraction
            when `include_top` is `False`.
            - `None` means that the output of the model will be the 4D tensor
                output of the last convolutional block.
            - `avg` means that global average pooling will be applied to the
                output of the last convolutional block, and thus the output of
                the model will be a 2D tensor.
            - `max` means that global max pooling will be applied.
        name: string, optional name to pass to the model, defaults to "{name}".
        classifier_activation: A `str` or callable. The activation function to
            use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top"
            layer.

    Returns:
      A `keras.Model` instance.
"""  # noqa: E501


def apply_mlp_block(x, mlp_dim, name=None):
    """An MLP block consisting of two linear layers with GELU activation in
    between.

    Args:
      x: input tensor.
      mlp_dim: integer, the number of units to be present in the first layer.
      name: string, block label.

    Returns:
      the updated input tensor.
    """
    if name is None:
        name = f"mlp_block_{backend.get_uid('mlp_block')}"

    y = layers.Dense(mlp_dim, name=f"{name}_dense_1")(x)
    y = layers.Activation("gelu", name=f"{name}_gelu")(y)
    return layers.Dense(x.shape[-1], name=f"{name}_dense_2")(y)


def apply_mixer_block(x, tokens_mlp_dim, channels_mlp_dim, name=None):
    """A mixer block.

    Args:
      x: input tensor.
      tokens_mlp_dim: integer, number of units to be present in the MLP block
        dealing with tokens.
      channels_mlp_dim: integer, number of units to be present in the MLP block
        dealing with channels.
      name: string, block label.

    Returns:
      the updated input tensor.
    """
    if name is None:
        name = f"mixer_block_{backend.get_uid('mlp_block')}"

    y = layers.LayerNormalization()(x)
    y = layers.Permute((2, 1))(y)

    y = apply_mlp_block(y, tokens_mlp_dim, name=f"{name}_token_mixing")
    y = layers.Permute((2, 1))(y)
    x = layers.Add()([x, y])

    y = layers.LayerNormalization()(x)
    y = apply_mlp_block(y, channels_mlp_dim, name=f"{name}_channel_mixing")
    return layers.Add()([x, y])


@keras.utils.register_keras_serializable(package="keras_cv.models")
class MLPMixer(keras.Model):
    """Instantiates the MLP Mixer architecture.

    Args:
      input_shape: tuple denoting the input shape, (224, 224, 3) for example.
      patch_size: integer denoting the size of the patches to be extracted
        from the inputs (16 for extracting 16x16 patches for example).
      num_blocks: integer, number of mixer blocks.
      hidden_dim: integer, dimension to which the patches will be linearly
        projected.
      tokens_mlp_dim: integer, dimension of the MLP block responsible for
        tokens.
      channels_mlp_dim: integer, dimension of the MLP block responsible for
        channels.
      include_rescaling: whether to rescale the inputs. If set to True,
        inputs will be passed through a `Rescaling(1/255.0)` layer.
      include_top: bool, whether to include the fully-connected
        layer at the top of the network. If provided, num_classes must be
        provided.
      num_classes: integer, optional number of classes to classify images
        into. Only to be specified if `include_top` is True.
      weights: one of `None` (random initialization) or a pretrained
        weight file path.
      input_tensor: optional Keras tensor (i.e., output of `layers.Input()`)
        to use as image input for the model.
      pooling: optional pooling mode for feature extraction
        when `include_top` is `False`.
        - `None` means that the output of the model will be
            the 4D tensor output of the
            last convolutional block.
        - `avg` means that global average pooling
            will be applied to the output of the
            last convolutional block, and thus
            the output of the model will be a 2D tensor.
        - `max` means that global max pooling will
            be applied.
      classifier_activation: A `str` or callable. The activation function to use
        on the "top" layer. Ignored unless `include_top=True`. Set
        `classifier_activation=None` to return the logits of the "top" layer.
        When loading pretrained weights, `classifier_activation` can only
        be `None` or `"softmax"`.
      name: string, optional name to pass to the model, defaults to "MLPMixer".

    Returns:
      A `keras.Model` instance.
    """

    def __init__(
        self,
        input_shape,
        patch_size,
        num_blocks,
        hidden_dim,
        tokens_mlp_dim,
        channels_mlp_dim,
        include_rescaling,
        include_top,
        num_classes=None,
        input_tensor=None,
        weights=None,
        pooling=None,
        classifier_activation="softmax",
        name="MLPMixer",
        **kwargs,
    ):
        if weights and not tf.io.gfile.exists(weights):
            raise ValueError(
                "The `weights` argument should be either "
                "`None` or the path to the weights file to be loaded. "
                f"Weights file not found at location: {weights}"
            )

        if include_top and not num_classes:
            raise ValueError(
                "If `include_top` is True, "
                "you should specify `num_classes`. "
                f"Received: num_classes={num_classes}"
            )

        if not isinstance(input_shape, tuple):
            raise ValueError("`input_shape` needs to be tuple.")

        if len(input_shape) != 3:
            raise ValueError(
                "`input_shape` needs to contain dimensions for three"
                " axes: height, width, and channel ((224, 224, 3) for example)."
            )

        if input_shape[0] != input_shape[1]:
            raise ValueError("Non-uniform resolutions are not supported.")

        if input_shape[0] % patch_size != 0:
            raise ValueError(
                "Input resolution should be divisible by the patch size."
            )

        inputs = utils.parse_model_inputs(input_shape, input_tensor)

        x = inputs
        if include_rescaling:
            x = layers.Rescaling(1 / 255.0)(x)

        x = layers.Conv2D(
            filters=hidden_dim,
            kernel_size=(patch_size, patch_size),
            strides=(patch_size, patch_size),
            padding="valid",
            name="patchify_and_projection",
        )(x)
        x = layers.Reshape((x.shape[1] * x.shape[2], x.shape[3]))(x)

        for i in range(num_blocks):
            x = apply_mixer_block(
                x, tokens_mlp_dim, channels_mlp_dim, name=f"mixer_block_{i}"
            )

        x = layers.LayerNormalization()(x)

        if include_top:
            x = layers.GlobalAveragePooling1D(name="avg_pool")(x)
            x = layers.Dense(
                num_classes,
                activation=classifier_activation,
                name="predictions",
            )(x)

        elif pooling == "avg":
            x = layers.GlobalAveragePooling1D(name="avg_pool")(x)
        elif pooling == "max":
            x = layers.GlobalMaxPooling1D(name="max_pool")(x)

        super().__init__(inputs=inputs, outputs=x, name=name, **kwargs)

        if weights is not None:
            self.load_weights(weights)

        self.patch_size = patch_size
        self.num_blocks = num_blocks
        self.hidden_dim = hidden_dim
        self.tokens_mlp_dim = tokens_mlp_dim
        self.channels_mlp_dim = channels_mlp_dim
        self.include_rescaling = include_rescaling
        self.include_top = include_top
        self.num_classes = num_classes
        self.input_tensor = input_tensor
        self.pooling = pooling
        self.classifier_activation = classifier_activation

    def get_config(self):
        return {
            "input_shape": self.input_shape[1:],
            "patch_size": self.patch_size,
            "num_blocks": self.num_blocks,
            "hidden_dim": self.hidden_dim,
            "tokens_mlp_dim": self.tokens_mlp_dim,
            "channels_mlp_dim": self.channels_mlp_dim,
            "include_rescaling": self.include_rescaling,
            "include_top": self.include_top,
            "num_classes": self.num_classes,
            "input_tensor": self.input_tensor,
            "pooling": self.pooling,
            "classifier_activation": self.classifier_activation,
            "name": self.name,
            "trainable": self.trainable,
        }

    @classmethod
    def from_config(cls, config):
        return cls(**config)


def MLPMixerB16(
    input_shape,
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    input_tensor=None,
    weights=None,
    pooling=None,
    name="MLPMixerB16",
    **kwargs,
):
    """Instantiates the MLPMixerB16 architecture."""

    return MLPMixer(
        input_shape=input_shape,
        patch_size=MODEL_CONFIGS["MLPMixerB16"]["patch_size"],
        num_blocks=MODEL_CONFIGS["MLPMixerB16"]["num_blocks"],
        hidden_dim=MODEL_CONFIGS["MLPMixerB16"]["hidden_dim"],
        tokens_mlp_dim=MODEL_CONFIGS["MLPMixerB16"]["tokens_mlp_dim"],
        channels_mlp_dim=MODEL_CONFIGS["MLPMixerB16"]["channels_mlp_dim"],
        include_rescaling=include_rescaling,
        include_top=include_top,
        num_classes=num_classes,
        input_tensor=input_tensor,
        weights=weights,
        pooling=pooling,
        name=name,
        **kwargs,
    )


def MLPMixerB32(
    input_shape,
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    input_tensor=None,
    weights=None,
    pooling=None,
    name="MLPMixerB32",
    **kwargs,
):
    """Instantiates the MLPMixerB32 architecture."""
    return MLPMixer(
        input_shape=input_shape,
        patch_size=MODEL_CONFIGS["MLPMixerB32"]["patch_size"],
        num_blocks=MODEL_CONFIGS["MLPMixerB32"]["num_blocks"],
        hidden_dim=MODEL_CONFIGS["MLPMixerB32"]["hidden_dim"],
        tokens_mlp_dim=MODEL_CONFIGS["MLPMixerB32"]["tokens_mlp_dim"],
        channels_mlp_dim=MODEL_CONFIGS["MLPMixerB32"]["channels_mlp_dim"],
        include_rescaling=include_rescaling,
        include_top=include_top,
        num_classes=num_classes,
        input_tensor=input_tensor,
        weights=weights,
        pooling=pooling,
        name=name,
        **kwargs,
    )


def MLPMixerL16(
    input_shape,
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    input_tensor=None,
    weights=None,
    pooling=None,
    name="MLPMixerL16",
    **kwargs,
):
    """Instantiates the MLPMixerL16 architecture."""
    return MLPMixer(
        input_shape=input_shape,
        patch_size=MODEL_CONFIGS["MLPMixerL16"]["patch_size"],
        num_blocks=MODEL_CONFIGS["MLPMixerL16"]["num_blocks"],
        hidden_dim=MODEL_CONFIGS["MLPMixerL16"]["hidden_dim"],
        tokens_mlp_dim=MODEL_CONFIGS["MLPMixerL16"]["tokens_mlp_dim"],
        channels_mlp_dim=MODEL_CONFIGS["MLPMixerL16"]["channels_mlp_dim"],
        include_rescaling=include_rescaling,
        include_top=include_top,
        num_classes=num_classes,
        input_tensor=input_tensor,
        weights=weights,
        pooling=pooling,
        name=name,
        **kwargs,
    )


setattr(MLPMixerB16, "__doc__", BASE_DOCSTRING.format(name="MLPMixerB16"))
setattr(MLPMixerB32, "__doc__", BASE_DOCSTRING.format(name="MLPMixerB32"))
setattr(MLPMixerL16, "__doc__", BASE_DOCSTRING.format(name="MLPMixerL16"))



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""SegFormer model preset configurations."""

from keras_cv.src.models.backbones.mix_transformer.mix_transformer_backbone_presets import (  # noqa: E501
    backbone_presets,
)

presets_no_weights = {
    "segformer_b0": {
        "metadata": {
            "description": ("SegFormer model with MiTB0 backbone."),
            "params": 3719027,
            "official_name": "SegFormerB0",
            "path": "segformer_b0",
        },
        "kaggle_handle": "kaggle://keras/segformer/keras/segformer_b0/2",
    },
    "segformer_b1": {
        "metadata": {
            "description": ("SegFormer model with MiTB1 backbone."),
            "params": 13682643,
            "official_name": "SegFormerB1",
            "path": "segformer_b1",
        },
        "kaggle_handle": "kaggle://keras/segformer/keras/segformer_b1/2",
    },
    "segformer_b2": {
        "metadata": {
            "description": ("SegFormer model with MiTB2 backbone."),
            "params": 24727507,
            "official_name": "SegFormerB2",
            "path": "segformer_b2",
        },
        "kaggle_handle": "kaggle://keras/segformer/keras/segformer_b2/2",
    },
    "segformer_b3": {
        "metadata": {
            "description": ("SegFormer model with MiTB3 backbone."),
            "params": 44603347,
            "official_name": "SegFormerB3",
            "path": "segformer_b3",
        },
        "kaggle_handle": "kaggle://keras/segformer/keras/segformer_b3/2",
    },
    "segformer_b4": {
        "metadata": {
            "description": ("SegFormer model with MiTB4 backbone."),
            "params": 61373907,
            "official_name": "SegFormerB4",
            "path": "segformer_b4",
        },
        "kaggle_handle": "kaggle://keras/segformer/keras/segformer_b4/2",
    },
    "segformer_b5": {
        "metadata": {
            "description": ("SegFormer model with MiTB5 backbone."),
            "params": 81974227,
            "official_name": "SegFormerB5",
            "path": "segformer_b5",
        },
        "kaggle_handle": "kaggle://keras/segformer/keras/segformer_b5/2",
    },
}

presets_with_weights = {
    "segformer_b0_imagenet": {
        "metadata": {
            "description": (
                "SegFormer model with a pretrained MiTB0 backbone."
            ),
            "params": 3719027,
            "official_name": "SegFormerB0",
            "path": "segformer_b0",
        },
        "kaggle_handle": "kaggle://keras/segformer/keras/segformer_b0_imagenet/2",  # noqa: E501
    },
}

presets = {
    **backbone_presets,  # Add MiTBackbone presets
    **presets_no_weights,
    **presets_with_weights,
}



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.models import utils
from keras_cv.src.models.backbones.backbone import Backbone
from keras_cv.src.models.object_detection.yolo_v8.yolo_v8_backbone_presets import (  # noqa: E501
    backbone_presets,
)
from keras_cv.src.models.object_detection.yolo_v8.yolo_v8_backbone_presets import (  # noqa: E501
    backbone_presets_with_weights,
)
from keras_cv.src.models.object_detection.yolo_v8.yolo_v8_layers import (
    apply_conv_bn,
)
from keras_cv.src.models.object_detection.yolo_v8.yolo_v8_layers import (
    apply_csp_block,
)
from keras_cv.src.utils.python_utils import classproperty


def apply_spatial_pyramid_pooling_fast(
    inputs, pool_size=5, activation="swish", name="spp_fast"
):
    channel_axis = -1
    input_channels = inputs.shape[channel_axis]
    hidden_channels = int(input_channels // 2)

    x = apply_conv_bn(
        inputs,
        hidden_channels,
        kernel_size=1,
        activation=activation,
        name=f"{name}_pre",
    )
    pool_1 = keras.layers.MaxPooling2D(
        pool_size=pool_size, strides=1, padding="same", name=f"{name}_pool1"
    )(x)
    pool_2 = keras.layers.MaxPooling2D(
        pool_size=pool_size, strides=1, padding="same", name=f"{name}_pool2"
    )(pool_1)
    pool_3 = keras.layers.MaxPooling2D(
        pool_size=pool_size, strides=1, padding="same", name=f"{name}_pool3"
    )(pool_2)

    out = ops.concatenate([x, pool_1, pool_2, pool_3], axis=channel_axis)
    out = apply_conv_bn(
        out,
        input_channels,
        kernel_size=1,
        activation=activation,
        name=f"{name}_output",
    )
    return out


@keras_cv_export("keras_cv.models.YOLOV8Backbone")
class YOLOV8Backbone(Backbone):
    """Implements the YOLOV8 backbone for object detection.

    This backbone is a variant of the `CSPDarkNetBackbone` architecture.

    For transfer learning use cases, make sure to read the
    [guide to transfer learning & fine-tuning](https://keras.io/guides/transfer_learning/).

    Args:
        stackwise_channels: A list of ints, the number of channels for each dark
            level in the model.
        stackwise_depth: A list of ints, the depth for each dark level in the
            model.
        include_rescaling: bool, whether to rescale the inputs. If set to
            True, inputs will be passed through a `Rescaling(1/255.0)` layer.
        activation: String. The activation functions to use in the backbone to
            use in the CSPDarkNet blocks. Defaults to "swish".
        input_shape: optional shape tuple, defaults to (None, None, 3).
        input_tensor: optional Keras tensor (i.e. output of `layers.Input()`)
            to use as image input for the model.

    Returns:
        A `keras.Model` instance.

    Examples:
    ```python
    input_data = tf.ones(shape=(8, 224, 224, 3))

    # Pretrained backbone
    model = keras_cv.models.YOLOV8Backbone.from_preset(
        "yolo_v8_xs_backbone_coco"
    )
    output = model(input_data)

    # Randomly initialized backbone with a custom config
    model = keras_cv.models.YOLOV8Backbone(
        stackwise_channels=[128, 256, 512, 1024],
        stackwise_depth=[3, 9, 9, 3],
        include_rescaling=False,
    )
    output = model(input_data)
    ```
    """  # noqa: E501

    def __init__(
        self,
        stackwise_channels,
        stackwise_depth,
        include_rescaling,
        activation="swish",
        input_shape=(None, None, 3),
        input_tensor=None,
        **kwargs,
    ):
        inputs = utils.parse_model_inputs(input_shape, input_tensor)

        x = inputs
        if include_rescaling:
            x = keras.layers.Rescaling(1 / 255.0)(x)

        """ Stem """
        stem_width = stackwise_channels[0]
        x = apply_conv_bn(
            x,
            stem_width // 2,
            kernel_size=3,
            strides=2,
            activation=activation,
            name="stem_1",
        )
        x = apply_conv_bn(
            x,
            stem_width,
            kernel_size=3,
            strides=2,
            activation=activation,
            name="stem_2",
        )

        """ blocks """
        pyramid_level_inputs = {"P1": utils.get_tensor_input_name(x)}
        for stack_id, (channel, depth) in enumerate(
            zip(stackwise_channels, stackwise_depth)
        ):
            stack_name = f"stack{stack_id + 1}"
            if stack_id >= 1:
                x = apply_conv_bn(
                    x,
                    channel,
                    kernel_size=3,
                    strides=2,
                    activation=activation,
                    name=f"{stack_name}_downsample",
                )
            x = apply_csp_block(
                x,
                depth=depth,
                expansion=0.5,
                activation=activation,
                name=f"{stack_name}_c2f",
            )

            if stack_id == len(stackwise_depth) - 1:
                x = apply_spatial_pyramid_pooling_fast(
                    x,
                    pool_size=5,
                    activation=activation,
                    name=f"{stack_name}_spp_fast",
                )
            pyramid_level_inputs[f"P{stack_id + 2}"] = (
                utils.get_tensor_input_name(x)
            )

        super().__init__(inputs=inputs, outputs=x, **kwargs)
        self.pyramid_level_inputs = pyramid_level_inputs
        self.stackwise_channels = stackwise_channels
        self.stackwise_depth = stackwise_depth
        self.include_rescaling = include_rescaling
        self.activation = activation

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "include_rescaling": self.include_rescaling,
                # Remove batch dimension from `input_shape`
                "input_shape": self.input_shape[1:],
                "stackwise_channels": self.stackwise_channels,
                "stackwise_depth": self.stackwise_depth,
                "activation": self.activation,
            }
        )
        return config

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return copy.deepcopy(backbone_presets)

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return copy.deepcopy(backbone_presets_with_weights)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops


@keras_cv_export(
    "keras_cv.models.retinanet.FeaturePyramid",
    package="keras_cv.models.retinanet",
)
class FeaturePyramid(keras.layers.Layer):
    """Builds the Feature Pyramid with the feature maps from the backbone."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.conv_c3_1x1 = keras.layers.Conv2D(256, 1, 1, "same")
        self.conv_c4_1x1 = keras.layers.Conv2D(256, 1, 1, "same")
        self.conv_c5_1x1 = keras.layers.Conv2D(256, 1, 1, "same")
        self.conv_c3_3x3 = keras.layers.Conv2D(256, 3, 1, "same")
        self.conv_c4_3x3 = keras.layers.Conv2D(256, 3, 1, "same")
        self.conv_c5_3x3 = keras.layers.Conv2D(256, 3, 1, "same")
        self.conv_c6_3x3 = keras.layers.Conv2D(256, 3, 2, "same")
        self.conv_c7_3x3 = keras.layers.Conv2D(256, 3, 2, "same")
        self.upsample_2x = keras.layers.UpSampling2D(2)

    def call(self, inputs, training=False):
        if isinstance(inputs, dict):
            c3_output = inputs["P3"]
            c4_output = inputs["P4"]
            c5_output = inputs["P5"]
        else:
            c3_output, c4_output, c5_output = inputs
        p3_output = self.conv_c3_1x1(c3_output, training=training)
        p4_output = self.conv_c4_1x1(c4_output, training=training)
        p5_output = self.conv_c5_1x1(c5_output, training=training)
        p4_output = p4_output + self.upsample_2x(p5_output, training=training)
        p3_output = p3_output + self.upsample_2x(p4_output, training=training)
        p3_output = self.conv_c3_3x3(p3_output, training=training)
        p4_output = self.conv_c4_3x3(p4_output, training=training)
        p5_output = self.conv_c5_3x3(p5_output, training=training)
        p6_output = self.conv_c6_3x3(c5_output, training=training)
        p7_output = self.conv_c7_3x3(ops.relu(p6_output), training=training)
        return p3_output, p4_output, p5_output, p6_output, p7_output

    def build(self, input_shape):
        p3_channels = input_shape["P3"][-1]
        p4_channels = input_shape["P4"][-1]
        p5_channels = input_shape["P5"][-1]
        self.conv_c3_1x1.build((None, None, None, p3_channels))
        self.conv_c4_1x1.build((None, None, None, p4_channels))
        self.conv_c5_1x1.build((None, None, None, p5_channels))
        self.conv_c3_3x3.build((None, None, None, 256))
        self.conv_c4_3x3.build((None, None, None, 256))
        self.conv_c5_3x3.build((None, None, None, 256))
        self.conv_c6_3x3.build((None, None, None, p5_channels))
        self.conv_c7_3x3.build((None, None, None, 256))
        self.built = True

    def compute_output_shape(self, input_shape):
        p3_shape = input_shape["P3"][:-1]
        p4_shape = input_shape["P4"][:-1]
        p5_shape = input_shape["P5"][:-1]
        return (
            (tuple(p3_shape) + (256,)),
            (tuple(p4_shape) + (256,)),
            (tuple(p5_shape) + (256,)),
            (tuple(p5_shape) + (256,)),
            (tuple(p5_shape) + (256,)),
        )



# Copyright 2024 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
from functools import partial

import numpy as np
from keras import layers

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.models import utils
from keras_cv.src.models.backbones.backbone import Backbone
from keras_cv.src.models.backbones.video_swin.video_swin_backbone_presets import (  # noqa: E501
    backbone_presets,
)
from keras_cv.src.models.backbones.video_swin.video_swin_backbone_presets import (  # noqa: E501
    backbone_presets_with_weights,
)
from keras_cv.src.models.backbones.video_swin.video_swin_layers import (
    VideoSwinBasicLayer,
)
from keras_cv.src.models.backbones.video_swin.video_swin_layers import (
    VideoSwinPatchingAndEmbedding,
)
from keras_cv.src.models.backbones.video_swin.video_swin_layers import (
    VideoSwinPatchMerging,
)
from keras_cv.src.utils.python_utils import classproperty


@keras_cv_export("keras_cv.models.VideoSwinBackbone", package="keras_cv.models")
class VideoSwinBackbone(Backbone):
    """A Video Swin Transformer backbone model.

    Args:
        input_shape (tuple[int], optional): The size of the input video in
            `(depth, height, width, channel)` format.
            Defaults to `(32, 224, 224, 3)`.
        input_tensor (KerasTensor, optional): Output of
            `keras.layers.Input()`) to use as video input for the model.
            Defaults to `None`.
        include_rescaling (bool, optional): Whether to rescale the inputs. If
            set to `True`, inputs will be passed through a `Rescaling(1/255.0)` layer
            and normalize with mean=[0.485, 0.456, 0.406] and std=[0.229, 0.224, 0.225].
            Defaults to `False`.
        patch_size (int | tuple(int)): The patch size for depth, height, and width
            dimensions respectively. Default: (2,4,4).
        embed_dim (int): Number of linear projection output channels.
            Default to 96.
        depths (tuple[int]): Depths of each Swin Transformer stage.
            Default to [2, 2, 6, 2]
        num_heads (tuple[int]): Number of attention head of each stage.
            Default to [3, 6, 12, 24]
        window_size (int): The window size for depth, height, and width
            dimensions respectively. Default to [8, 7, 7].
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
            Default to 4.
        qkv_bias (bool): If True, add a learnable bias to query, key, value.
            Default to True.
        qk_scale (float): Override default qk scale of head_dim ** -0.5 if set.
            Default to None.
        drop_rate (float): Float between 0 and 1. Fraction of the input units to drop.
            Default: 0.
        attn_drop_rate (float): Float between 0 and 1. Attention dropout rate.
            Default: 0.
        drop_path_rate (float): Float between 0 and 1. Stochastic depth rate.
            Default: 0.2.
        patch_norm (bool): If True, add layer normalization after patch embedding.
            Default to False.

    Example:
    ```python
    # Build video swin backbone without top layer
    model = VideoSwinSBackbone(
        include_rescaling=True, input_shape=(8, 256, 256, 3),
    )
    videos = keras.ops.ones((1, 8, 256, 256, 3))
    outputs = model.predict(videos)
    ```

    References:
        - [Video Swin Transformer](https://arxiv.org/abs/2106.13230)
        - [Official Code](https://github.com/SwinTransformer/Video-Swin-Transformer)
    """  # noqa: E501

    def __init__(
        self,
        *,
        include_rescaling=False,
        input_shape=(32, 224, 224, 3),
        input_tensor=None,
        embed_dim=96,
        patch_size=[2, 4, 4],
        window_size=[8, 7, 7],
        mlp_ratio=4.0,
        patch_norm=True,
        drop_rate=0.0,
        attn_drop_rate=0.0,
        drop_path_rate=0.2,
        depths=[2, 2, 6, 2],
        num_heads=[3, 6, 12, 24],
        qkv_bias=True,
        qk_scale=None,
        **kwargs,
    ):
        # Parse input specification.
        input_spec = utils.parse_model_inputs(
            input_shape, input_tensor, name="videos"
        )

        # Check that the input video is well specified.
        if (
            input_spec.shape[-4] is None
            or input_spec.shape[-3] is None
            or input_spec.shape[-2] is None
        ):
            raise ValueError(
                "Depth, height and width of the video must be specified"
                " in `input_shape`."
            )

        x = input_spec

        if include_rescaling:
            # Use common rescaling strategy across keras_cv
            x = keras.layers.Rescaling(1.0 / 255.0)(x)

            # VideoSwin scales inputs based on the ImageNet mean/stddev.
            # Officially, Videw Swin takes tensor of [0-255] ranges.
            # And use mean=[123.675, 116.28, 103.53] and
            # std=[58.395, 57.12, 57.375] for normalization.
            # So, if include_rescaling is set to True, then, to match with the
            # official scores, following normalization should be added.
            x = layers.Normalization(
                mean=[0.485, 0.456, 0.406],
                variance=[0.229**2, 0.224**2, 0.225**2],
            )(x)

        norm_layer = partial(layers.LayerNormalization, epsilon=1e-05)

        x = VideoSwinPatchingAndEmbedding(
            patch_size=patch_size,
            embed_dim=embed_dim,
            norm_layer=norm_layer if patch_norm else None,
            name="videoswin_patching_and_embedding",
        )(x)
        x = layers.Dropout(drop_rate, name="pos_drop")(x)

        dpr = np.linspace(0.0, drop_path_rate, sum(depths)).tolist()
        num_layers = len(depths)
        for i in range(num_layers):
            layer = VideoSwinBasicLayer(
                input_dim=int(embed_dim * 2**i),
                depth=depths[i],
                num_heads=num_heads[i],
                window_size=window_size,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop_rate=drop_rate,
                attn_drop_rate=attn_drop_rate,
                drop_path_rate=dpr[sum(depths[:i]) : sum(depths[: i + 1])],
                norm_layer=norm_layer,
                downsampling_layer=(
                    VideoSwinPatchMerging if (i < num_layers - 1) else None
                ),
                name=f"videoswin_basic_layer_{i + 1}",
            )
            x = layer(x)

        x = norm_layer(axis=-1, epsilon=1e-05, name="videoswin_top_norm")(x)
        super().__init__(inputs=input_spec, outputs=x, **kwargs)

        self.include_rescaling = include_rescaling
        self.input_tensor = input_tensor
        self.embed_dim = embed_dim
        self.patch_size = patch_size
        self.window_size = window_size
        self.mlp_ratio = mlp_ratio
        self.norm_layer = norm_layer
        self.patch_norm = patch_norm
        self.drop_rate = drop_rate
        self.attn_drop_rate = attn_drop_rate
        self.drop_path_rate = drop_path_rate
        self.num_layers = len(depths)
        self.num_heads = num_heads
        self.qkv_bias = qkv_bias
        self.qk_scale = qk_scale
        self.depths = depths

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "include_rescaling": self.include_rescaling,
                "input_shape": self.input_shape[1:],
                "input_tensor": self.input_tensor,
                "embed_dim": self.embed_dim,
                "patch_norm": self.patch_norm,
                "window_size": self.window_size,
                "patch_size": self.patch_size,
                "mlp_ratio": self.mlp_ratio,
                "drop_rate": self.drop_rate,
                "drop_path_rate": self.drop_path_rate,
                "attn_drop_rate": self.attn_drop_rate,
                "depths": self.depths,
                "num_heads": self.num_heads,
                "qkv_bias": self.qkv_bias,
                "qk_scale": self.qk_scale,
            }
        )
        return config

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return copy.deepcopy(backbone_presets)

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return copy.deepcopy(backbone_presets_with_weights)

    @property
    def pyramid_level_inputs(self):
        raise NotImplementedError(
            "The `VideoSwinBackbone` model doesn't compute"
            " pyramid level features."
        )



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.layers.preprocessing.vectorized_base_image_augmentation_layer import (  # noqa: E501
    VectorizedBaseImageAugmentationLayer,
)
from keras_cv.src.utils import preprocessing as preprocessing_utils


@keras_cv_export("keras_cv.layers.RandomHue")
class RandomHue(VectorizedBaseImageAugmentationLayer):
    """Randomly adjusts the hue on given images.

    This layer will randomly increase/reduce the hue for the input RGB
    images.

    The image hue is adjusted by converting the image(s) to HSV and rotating the
    hue channel (H) by delta. The image is then converted back to RGB.

    Args:
        factor: A tuple of two floats, a single float or
            `keras_cv.FactorSampler`. `factor` controls the extent to which the
            image hue is impacted. `factor=0.0` makes this layer perform a
            no-op operation, while a value of 1.0 performs the most aggressive
            contrast adjustment available. If a tuple is used, a `factor` is
            sampled between the two values for every image augmented. If a
            single float is used, a value between `0.0` and the passed float is
            sampled. In order to ensure the value is always the same, please
            pass a tuple with two identical floats: `(0.5, 0.5)`.
        value_range: the range of values the incoming images will have.
            Represented as a two number tuple written [low, high]. This is
            typically either `[0, 1]` or `[0, 255]` depending on how your
            preprocessing pipeline is set up.
        seed: Integer. Used to create a random seed.

    Example:
    ```python
    (images, labels), _ = keras.datasets.cifar10.load_data()
    random_hue = keras_cv.layers.preprocessing.RandomHue()
    augmented_images = random_hue(images)
    ```
    """

    def __init__(self, factor, value_range, seed=None, **kwargs):
        super().__init__(seed=seed, **kwargs)
        self.factor = preprocessing_utils.parse_factor(
            factor,
        )
        self.value_range = value_range
        self.seed = seed

    def get_random_transformation_batch(self, batch_size, **kwargs):
        invert = self._random_generator.uniform((batch_size,), 0, 1, tf.float32)
        invert = tf.where(
            invert > 0.5, -tf.ones_like(invert), tf.ones_like(invert)
        )
        # We must scale self.factor() to the range [-0.5, 0.5]. This is because
        # the tf.image operation performs rotation on the hue saturation value
        # orientation. This can be thought of as an angle in the range
        # [-180, 180]
        return invert * self.factor(shape=(batch_size,)) * 0.5

    def augment_ragged_image(self, image, transformation, **kwargs):
        return self.augment_images(
            images=image, transformations=transformation, **kwargs
        )

    def augment_images(self, images, transformations, **kwargs):
        images = preprocessing_utils.transform_value_range(
            images, self.value_range, (0, 1), dtype=self.compute_dtype
        )
        adjust_factors = tf.cast(transformations, images.dtype)
        # broadcast
        adjust_factors = adjust_factors[..., tf.newaxis, tf.newaxis]

        # tf.image.adjust_hue expects floats to be in range [0, 1]
        images = tf.image.rgb_to_hsv(images)
        h_channel = images[..., 0] + adjust_factors
        h_channel = tf.where(h_channel > 1.0, h_channel - 1.0, h_channel)
        h_channel = tf.where(h_channel < 0.0, h_channel + 1.0, h_channel)
        images = tf.stack([h_channel, images[..., 1], images[..., 2]], axis=-1)
        images = tf.image.hsv_to_rgb(images)
        # RandomHue is one of the rare KPLs that needs to clip
        images = tf.clip_by_value(images, 0, 1)
        images = preprocessing_utils.transform_value_range(
            images, (0, 1), self.value_range, dtype=self.compute_dtype
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def get_config(self):
        config = {
            "factor": self.factor,
            "value_range": self.value_range,
            "seed": self.seed,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))

    @classmethod
    def from_config(cls, config):
        if isinstance(config["factor"], dict):
            config["factor"] = keras.utils.deserialize_keras_object(
                config["factor"]
            )
        return cls(**config)



# Copyright 2022 The KerasCV Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""IoU3D using a custom TF op."""

from keras_cv.src.utils.resource_loader import LazySO

keras_cv_custom_ops = LazySO("custom_ops/_keras_cv_custom_ops.so")


def iou_3d(y_true, y_pred):
    """Implements IoU computation for 3D upright rotated bounding boxes.

    Note that this is implemented using a custom TensorFlow op. If you don't
    have KerasCV installed with custom ops, calling this will fail.

    Boxes should have the format CENTER_XYZ_DXDYDZ_PHI. Refer to
    https://github.com/keras-team/keras-cv/blob/master/keras_cv/bounding_box_3d/formats.py
    for more details on supported bounding box formats.

    Example:
    ```python
    y_true = [[0, 0, 0, 2, 2, 2, 0], [1, 1, 1, 2, 2, 2, 3 * math.pi / 4]]
    y_pred = [[1, 1, 1, 2, 2, 2, math.pi / 4], [1, 1, 1, 2, 2, 2, 0]]
    iou_3d(y_true, y_pred)
    ```
    """

    return keras_cv_custom_ops.ops.kcv_pairwise_iou3d(y_true, y_pred)



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.layers.preprocessing.vectorized_base_image_augmentation_layer import (  # noqa: E501
    VectorizedBaseImageAugmentationLayer,
)
from keras_cv.src.utils import preprocessing as preprocessing_utils


@keras_cv_export("keras_cv.layers.RandomSaturation")
class RandomSaturation(VectorizedBaseImageAugmentationLayer):
    """Randomly adjusts the saturation on given images.

    This layer will randomly increase/reduce the saturation for the input RGB
    images.

    Args:
        factor: A tuple of two floats, a single float or
            `keras_cv.FactorSampler`. `factor` controls the extent to which the
            image saturation is impacted. `factor=0.5` makes this layer perform
            a no-op operation. `factor=0.0` makes the image to be fully
            grayscale. `factor=1.0` makes the image to be fully saturated.
            Values should be between `0.0` and `1.0`. If a tuple is used, a
            `factor` is sampled between the two values for every image
            augmented. If a single float is used, a value between `0.0` and the
            passed float is sampled. In order to ensure the value is always the
            same, please pass a tuple with two identical floats: `(0.5, 0.5)`.
        seed: Integer. Used to create a random seed.

    Example:
    ```python
    (images, labels), _ = keras.datasets.cifar10.load_data()
    random_saturation = keras_cv.layers.preprocessing.RandomSaturation()
    augmented_images = random_saturation(images)
    ```
    """

    def __init__(self, factor, seed=None, **kwargs):
        super().__init__(seed=seed, **kwargs)
        self.factor = preprocessing_utils.parse_factor(
            factor,
            min_value=0.0,
            max_value=1.0,
        )
        self.seed = seed

    def get_random_transformation_batch(self, batch_size, **kwargs):
        return self.factor(shape=(batch_size,))

    def augment_ragged_image(self, image, transformation, **kwargs):
        return self.augment_images(
            images=image, transformations=transformation, **kwargs
        )

    def augment_images(self, images, transformations, **kwargs):
        # Convert the factor range from [0, 1] to [0, +inf]. Note that the
        # tf.image.adjust_saturation is trying to apply the following math
        # formula `output_saturation = input_saturation * factor`. We use the
        # following method to the do the mapping.
        # `y = x / (1 - x)`.
        # This will ensure:
        #   y = +inf when x = 1 (full saturation)
        #   y = 1 when x = 0.5 (no augmentation)
        #   y = 0 when x = 0 (full gray scale)

        # Convert the transformation to tensor in case it is a float. When
        # transformation is 1.0, then it will result in to divide by zero error,
        # but it will be handled correctly when it is a one tensor.
        transformations = tf.convert_to_tensor(transformations)
        adjust_factors = transformations / (1 - transformations)
        adjust_factors = tf.cast(adjust_factors, dtype=images.dtype)

        images = tf.image.rgb_to_hsv(images)
        s_channel = tf.multiply(
            images[..., 1], adjust_factors[..., tf.newaxis, tf.newaxis]
        )
        s_channel = tf.clip_by_value(
            s_channel, clip_value_min=0.0, clip_value_max=1.0
        )
        images = tf.stack([images[..., 0], s_channel, images[..., 2]], axis=-1)
        images = tf.image.hsv_to_rgb(images)
        return images

    def augment_bounding_boxes(
        self, bounding_boxes, transformation=None, **kwargs
    ):
        return bounding_boxes

    def augment_labels(self, labels, transformations=None, **kwargs):
        return labels

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def get_config(self):
        config = {
            "factor": self.factor,
            "seed": self.seed,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))

    @classmethod
    def from_config(cls, config):
        if isinstance(config["factor"], dict):
            config["factor"] = keras.utils.deserialize_keras_object(
                config["factor"]
            )
        return cls(**config)



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities with miscellaneous python extensions."""


class classproperty(property):
    """Define a class level property."""

    def __get__(self, _, owner_cls):
        return self.fget(owner_cls)


def format_docstring(**replacements):
    """Format a python docstring using a dictionary of replacements.

    This decorator can be placed on a function, class or method to format it's
    docstring with python variables.

    The decorator will replace any double bracketed variable with a kwargs
    value passed to the decorator itself. For example
    `@format_docstring(name="foo")` will replace any occurrence of `{{name}}` in
    the docstring with the string literal `foo`.
    """

    def decorate(obj):
        doc = obj.__doc__
        # We use `str.format()` to replace variables in the docstring, but use
        # double brackets, e.g. {{var}}, to mark format strings. So we need to
        # to swap all double and single brackets in the source docstring.
        doc = "{".join(part.replace("{", "{{") for part in doc.split("{{"))
        doc = "}".join(part.replace("}", "}}") for part in doc.split("}}"))
        obj.__doc__ = doc.format(**replacements)
        return obj

    return decorate



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

import os

# Import everything from /api/ into keras.
from keras_cv.api import *  # noqa: F403
from keras_cv.api import __version__  # Import * ignores names start with "_".

# Add everything in /api/ to the module search path.
__path__.append(os.path.join(os.path.dirname(__file__), "api"))  # noqa: F405

# Don't pollute namespace.
del os


# Never autocomplete `.src` or `.api` on an imported keras object.
def __dir__():
    keys = dict.fromkeys((globals().keys()))
    keys.pop("src")
    keys.pop("api")
    return list(keys)


# Don't import `.src` or `.api` during `from keras import *`.
__all__ = [
    name
    for name in globals().keys()
    if not (name.startswith("_") or name in ("src", "api"))
]



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Converter functions for working with keypoints formats."""

import tensorflow as tf

from keras_cv.src.api_export import keras_cv_export


# Internal exception
class _RequiresImagesException(Exception):
    pass


def _rel_xy_to_xy(keypoints, images=None):
    if images is None:
        raise _RequiresImagesException()
    shape = tf.cast(tf.shape(images), keypoints.dtype)
    h, w = shape[1], shape[2]

    x, y, rest = tf.split(keypoints, [1, 1, keypoints.shape[-1] - 2], axis=-1)

    return tf.concat([x * w, y * h, rest], axis=-1)


def _xy_to_rel_xy(keypoints, images=None):
    if images is None:
        raise _RequiresImagesException()
    shape = tf.cast(tf.shape(images), keypoints.dtype)
    h, w = shape[1], shape[2]

    x, y, rest = tf.split(keypoints, [1, 1, keypoints.shape[-1] - 2], axis=-1)

    return tf.concat([x / w, y / h, rest], axis=-1)


def _xy_noop(keypoints, images=None):
    return keypoints


TO_XY_CONVERTERS = {
    "xy": _xy_noop,
    "rel_xy": _rel_xy_to_xy,
}

FROM_XY_CONVERTERS = {
    "xy": _xy_noop,
    "rel_xy": _xy_to_rel_xy,
}


@keras_cv_export(
    "keras_cv.keypoint.convert_format", package="keras_cv.keypoint"
)
def convert_format(keypoints, source, target, images=None, dtype=None):
    """Converts keypoints from one format to another.

    Supported formats are:
    - `"xy"`, absolute pixel positions.
    - `"rel_xyxy"`. relative pixel positions.

    Formats are case-insensitive. It is recommended that you
    capitalize width and height to maximize the visual difference
    between `"xyWH"` and `"xyxy"`.

    Relative formats, abbreviated `rel`, make use of the shapes of the
    `images` passed. In these formats, the coordinates, widths, and
    heights are all specified as percentages of the host image.
    `images` may be a ragged Tensor. Note that using a ragged Tensor
    for images may cause a substantial performance loss, as each image
    will need to be processed separately due to the mismatching image
    shapes.

    Example:

    ```python
    images, keypoints = load_my_dataset()
    keypoints_in_rel = keras_cv.keypoint.convert_format(
        keypoint,
        source='xy',
        target='rel_xy',
        images=images,
    )
    ```

    Args:
        keypoints: tf.Tensor or tf.RaggedTensor representing keypoints
            in the format specified in the `source` parameter.
            `keypoints` can optionally have extra dimensions stacked
            on the final axis to store metadata. keypoints should
            have a rank between 2 and 4, with the shape
            `[num_boxes,*]`, `[batch_size, num_boxes, *]` or
            `[batch_size, num_groups, num_keypoints,*]`.
        source: One of {" ".join([f'"{f}"' for f in
            TO_XY_CONVERTERS.keys()])}. Used to specify the original
            format of the `boxes` parameter.
        target: One of {" ".join([f'"{f}"' for f in
            TO_XY_CONVERTERS.keys()])}. Used to specify the
            destination format of the `boxes` parameter.
        images: (Optional) a batch of images aligned with `boxes` on
            the first axis. Should be rank 3 (`HWC` format) or 4
            (`BHWC` format). Used in some converters to compute
            relative pixel values of the bounding box dimensions.
            Required when transforming from a rel format to a non-rel
            format.
        dtype: the data type to use when transforming the boxes.
            Defaults to None, i.e. `keypoints` dtype.
    """

    source = source.lower()
    target = target.lower()
    if source not in TO_XY_CONVERTERS:
        raise ValueError(
            f"convert_format() received an unsupported format for the argument "
            f"`source`. `source` should be one of {TO_XY_CONVERTERS.keys()}. "
            f"Got source={source}"
        )
    if target not in FROM_XY_CONVERTERS:
        raise ValueError(
            f"convert_format() received an unsupported format for the argument "
            f"`target`. `target` should be one of {FROM_XY_CONVERTERS.keys()}. "
            f"Got target={target}"
        )

    if dtype:
        keypoints = tf.cast(keypoints, dtype)

    if source == target:
        return keypoints

    keypoints, images, squeeze_axis = _format_inputs(keypoints, images)

    try:
        in_xy = TO_XY_CONVERTERS[source](keypoints, images=images)
        result = FROM_XY_CONVERTERS[target](in_xy, images=images)
    except _RequiresImagesException:
        raise ValueError(
            "convert_format() must receive `images` when transforming "
            f"between relative and absolute formats. "
            f"convert_format() received source=`{source}`, target=`{target}`, "
            f"but images={images}"
        )

    return _format_outputs(result, squeeze_axis)


def _format_inputs(keypoints, images):
    keypoints_rank = len(keypoints.shape)
    if keypoints_rank > 4:
        raise ValueError(
            "Expected keypoints rank to be in [2, 4], got "
            f"len(keypoints.shape)={keypoints_rank}."
        )
    keypoints_includes_batch = keypoints_rank > 2
    keypoints_are_grouped = keypoints_rank == 4
    if images is not None:
        images_rank = len(images.shape)
        if images_rank > 4 or images_rank < 3:
            raise ValueError(
                "Expected images rank to be 3 or 4, got "
                f"len(images.shape)={images_rank}."
            )
        images_include_batch = images_rank == 4
        if keypoints_includes_batch != images_include_batch:
            raise ValueError(
                "convert_format() expects both `keypoints` and `images` to be "
                "batched or both unbatched. Received "
                f"len(keypoints.shape)={keypoints_rank}, "
                f"len(images.shape)={images_rank}. Expected either "
                "len(keypoints.shape)=2 and len(images.shape)=3, or "
                "len(keypoints.shape)>=3 and len(images.shape)=4."
            )
        if not images_include_batch:
            images = tf.expand_dims(images, axis=0)

    squeeze_axis = []
    if not keypoints_includes_batch:
        keypoints = tf.expand_dims(keypoints, axis=0)
        squeeze_axis.append(0)
    if not keypoints_are_grouped:
        keypoints = tf.expand_dims(keypoints, axis=1)
        squeeze_axis.append(1)

    return keypoints, images, squeeze_axis


def _format_outputs(result, squeeze_axis):
    if len(squeeze_axis) == 0:
        return result
    return tf.squeeze(result, axis=squeeze_axis)



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.layers.vit_det_layers import AddPositionalEmbedding
from keras_cv.src.layers.vit_det_layers import ViTDetPatchingAndEmbedding
from keras_cv.src.layers.vit_det_layers import WindowedTransformerEncoder
from keras_cv.src.models import utils
from keras_cv.src.models.backbones.backbone import Backbone
from keras_cv.src.models.backbones.vit_det.vit_det_backbone_presets import (
    backbone_presets,
)
from keras_cv.src.models.backbones.vit_det.vit_det_backbone_presets import (
    backbone_presets_with_weights,
)
from keras_cv.src.utils.python_utils import classproperty


@keras_cv_export("keras_cv.models.ViTDetBackbone", package="keras_cv.models")
class ViTDetBackbone(Backbone):
    """A ViT image encoder that uses a windowed transformer encoder and
    relative positional encodings.

    Args:
        input_shape (tuple[int], optional): The size of the input image in
            `(H, W, C)` format. Defaults to `(1024, 1024, 3)`.
        input_tensor (KerasTensor, optional): Output of
            `keras.layers.Input()`) to use as image input for the model.
            Defaults to `None`.
        include_rescaling (bool, optional): Whether to rescale the inputs. If
            set to `True`, inputs will be passed through a
            `Rescaling(1/255.0)` layer. Defaults to `False`.
        patch_size (int, optional): the patch size to be supplied to the
            Patching layer to turn input images into a flattened sequence of
            patches. Defaults to `16`.
        embed_dim (int, optional): The latent dimensionality to be projected
            into in the output of each stacked windowed transformer encoder.
            Defaults to `768`.
        depth (int, optional): The number of transformer encoder layers to
            stack in the Vision Transformer. Defaults to `12`.
        mlp_dim (int, optional): The dimensionality of the hidden Dense
            layer in the transformer MLP head. Defaults to `768*4`.
        num_heads (int, optional): the number of heads to use in the
            `MultiHeadAttentionWithRelativePE` layer of each transformer
            encoder. Defaults to `12`.
        out_chans (int, optional): The number of channels (features) in the
            output (image encodings). Defaults to `256`.
        use_bias (bool, optional): Whether to use bias to project the keys,
            queries, and values in the attention layer. Defaults to `True`.
        use_abs_pos (bool, optional): Whether to add absolute positional
            embeddings to the output patches. Defaults to `True`.
        use_rel_pos (bool, optional): Whether to use relative positional
            emcodings in the attention layer. Defaults to `True`.
        window_size (int, optional): The size of the window for windowed
            attention in the transformer encoder blocks. Defaults to `14`.
        global_attention_indices (list, optional): Indexes for blocks using
            global attention. Defaults to `[2, 5, 8, 11]`.
        layer_norm_epsilon (int, optional): The epsilon to use in the layer
            normalization blocks in transformer encoder. Defaults to `1e-6`.

    References:
        - [Segment Anything paper](https://arxiv.org/abs/2304.02643)
        - [Segment Anything GitHub](https://github.com/facebookresearch/segment-anything)
        - [Detectron2](https://github.com/facebookresearch/detectron2)
    """  # noqa: E501

    def __init__(
        self,
        *,
        include_rescaling,
        input_shape=(1024, 1024, 3),
        input_tensor=None,
        patch_size=16,
        embed_dim=768,
        depth=12,
        mlp_dim=768 * 4,
        num_heads=12,
        out_chans=256,
        use_bias=True,
        use_abs_pos=True,
        use_rel_pos=True,
        window_size=14,
        global_attention_indices=[2, 5, 8, 11],
        layer_norm_epsilon=1e-6,
        **kwargs
    ):
        img_input = utils.parse_model_inputs(
            input_shape, input_tensor, name="images"
        )

        # Check that the input image is well specified.
        if img_input.shape[-3] is None or img_input.shape[-2] is None:
            raise ValueError(
                "Height and width of the image must be specified"
                " in `input_shape`."
            )
        if img_input.shape[-3] != img_input.shape[-2]:
            raise ValueError(
                "Input image must be square i.e. the height must"
                " be equal to the width in the `input_shape`"
                " tuple/tensor."
            )

        img_size = img_input.shape[-3]

        x = img_input

        if include_rescaling:
            # Use common rescaling strategy across keras_cv
            x = keras.layers.Rescaling(1.0 / 255.0)(x)

        # VITDet scales inputs based on the standard ImageNet mean/stddev.
        x = (x - ops.array([0.485, 0.456, 0.406], dtype=x.dtype)) / (
            ops.array([0.229, 0.224, 0.225], dtype=x.dtype)
        )

        x = ViTDetPatchingAndEmbedding(
            kernel_size=(patch_size, patch_size),
            strides=(patch_size, patch_size),
            embed_dim=embed_dim,
        )(x)
        if use_abs_pos:
            x = AddPositionalEmbedding(img_size, patch_size, embed_dim)(x)

        for i in range(depth):
            x = WindowedTransformerEncoder(
                project_dim=embed_dim,
                mlp_dim=mlp_dim,
                num_heads=num_heads,
                use_bias=use_bias,
                use_rel_pos=use_rel_pos,
                window_size=(
                    window_size if i not in global_attention_indices else 0
                ),
                input_size=(img_size // patch_size, img_size // patch_size),
            )(x)
        x = keras.models.Sequential(
            [
                keras.layers.Conv2D(
                    filters=out_chans, kernel_size=1, use_bias=False
                ),
                keras.layers.LayerNormalization(epsilon=1e-6),
                keras.layers.Conv2D(
                    filters=out_chans,
                    kernel_size=3,
                    padding="same",
                    use_bias=False,
                ),
                keras.layers.LayerNormalization(epsilon=1e-6),
            ]
        )(x)

        super().__init__(inputs=img_input, outputs=x, **kwargs)

        self.patch_size = patch_size
        self.embed_dim = embed_dim
        self.depth = depth
        self.mlp_dim = mlp_dim
        self.num_heads = num_heads
        self.out_chans = out_chans
        self.use_bias = use_bias
        self.use_rel_pos = use_rel_pos
        self.use_abs_pos = use_abs_pos
        self.window_size = window_size
        self.global_attention_indices = global_attention_indices
        self.layer_norm_epsilon = layer_norm_epsilon
        self.input_tensor = input_tensor
        self.include_rescaling = include_rescaling

    @property
    def pyramid_level_inputs(self):
        raise NotImplementedError(
            "The `ViTDetBackbone` model doesn't compute"
            " pyramid level features."
        )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "input_shape": self.input_shape[1:],
                "input_tensor": self.input_tensor,
                "include_rescaling": self.include_rescaling,
                "patch_size": self.patch_size,
                "embed_dim": self.embed_dim,
                "depth": self.depth,
                "mlp_dim": self.mlp_dim,
                "num_heads": self.num_heads,
                "out_chans": self.out_chans,
                "use_bias": self.use_bias,
                "use_abs_pos": self.use_abs_pos,
                "use_rel_pos": self.use_rel_pos,
                "window_size": self.window_size,
                "global_attention_indices": self.global_attention_indices,
                "layer_norm_epsilon": self.layer_norm_epsilon,
            }
        )
        return config

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return copy.deepcopy(backbone_presets)

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return copy.deepcopy(backbone_presets_with_weights)



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops


@keras_cv_export("keras_cv.layers.SegFormerMultiheadAttention")
class SegFormerMultiheadAttention(keras.layers.Layer):
    def __init__(self, project_dim, num_heads, sr_ratio):
        """
        Efficient MultiHeadAttention implementation as a Keras layer.
        A huge bottleneck in scaling transformers is the self-attention layer
        with an O(n^2) complexity.

        SegFormerMultiheadAttention performs a sequence reduction (SR) operation
        with a given ratio, to reduce the sequence length before performing key and value projections,
        reducing the O(n^2) complexity to O(n^2/R) where R is the sequence reduction ratio.

        References:
        - [SegFormer: Simple and Efficient Design for Semantic Segmentation with Transformers](https://arxiv.org/abs/2105.15203) (CVPR 2021) # noqa: E501
        - [NVlabs' official implementation](https://github.com/NVlabs/SegFormer/blob/master/mmseg/models/backbones/mix_transformer.py) # noqa: E501
        - [@sithu31296's reimplementation](https://github.com/sithu31296/semantic-segmentation/blob/main/semseg/models/backbones/mit.py) # noqa: E501
        - [Ported from the TensorFlow implementation from DeepVision](https://github.com/DavidLandup0/deepvision/blob/main/deepvision/layers/efficient_attention.py) # noqa: E501

        Args:
            project_dim: integer, the dimensionality of the projection
                of the `SegFormerMultiheadAttention` layer.
            num_heads: integer, the number of heads to use in the
                attention computation.
            sr_ratio: integer, the sequence reduction ratio to perform
                on the sequence before key and value projections.

        Example:

        ```
        tensor = tf.random.uniform([1, 196, 32])
        output = keras_cv.layers.SegFormerMultiheadAttention(project_dim=768,
                                                            num_heads=2,
                                                            sr_ratio=4)(tensor)
        print(output.shape) # (1, 196, 32)
        ```
        """
        super().__init__()
        self.num_heads = num_heads
        self.sr_ratio = sr_ratio
        self.scale = (project_dim // num_heads) ** -0.5
        self.q = keras.layers.Dense(project_dim)
        self.k = keras.layers.Dense(project_dim)
        self.v = keras.layers.Dense(project_dim)
        self.proj = keras.layers.Dense(project_dim)

        if sr_ratio > 1:
            self.sr = keras.layers.Conv2D(
                filters=project_dim,
                kernel_size=sr_ratio,
                strides=sr_ratio,
                padding="same",
            )
            self.norm = keras.layers.LayerNormalization()

    def call(self, x):
        input_shape = ops.shape(x)
        H, W = int(math.sqrt(input_shape[1])), int(math.sqrt(input_shape[1]))
        B, C = input_shape[0], input_shape[2]

        q = self.q(x)
        q = ops.reshape(
            q,
            (
                input_shape[0],
                input_shape[1],
                self.num_heads,
                input_shape[2] // self.num_heads,
            ),
        )
        q = ops.transpose(q, [0, 2, 1, 3])

        if self.sr_ratio > 1:
            x = ops.reshape(
                ops.transpose(x, [0, 2, 1]),
                (B, H, W, C),
            )
            x = self.sr(x)
            x = ops.reshape(x, [input_shape[0], input_shape[2], -1])
            x = ops.transpose(x, [0, 2, 1])
            x = self.norm(x)

        k = self.k(x)
        v = self.v(x)

        k = ops.transpose(
            ops.reshape(
                k,
                [B, -1, self.num_heads, C // self.num_heads],
            ),
            [0, 2, 1, 3],
        )

        v = ops.transpose(
            ops.reshape(
                v,
                [B, -1, self.num_heads, C // self.num_heads],
            ),
            [0, 2, 1, 3],
        )

        attn = (q @ ops.transpose(k, [0, 1, 3, 2])) * self.scale
        attn = ops.nn.softmax(attn, axis=-1)

        attn = attn @ v
        attn = ops.reshape(
            ops.transpose(attn, [0, 2, 1, 3]),
            [input_shape[0], input_shape[1], input_shape[2]],
        )

        x = self.proj(attn)
        return x



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import numpy as np

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.models.backbones.backbone_presets import backbone_presets
from keras_cv.src.models.backbones.backbone_presets import (
    backbone_presets_with_weights,
)
from keras_cv.src.models.segmentation.segment_anything.sam_presets import (
    sam_presets,
)
from keras_cv.src.models.task import Task
from keras_cv.src.utils.python_utils import classproperty


@keras_cv_export(
    [
        "keras_cv.models.SegmentAnythingModel",
        "keras_cv.models.segmentation.SegmentAnythingModel",
    ],
    package="keras_cv.models",
)
class SegmentAnythingModel(Task):
    """
    The Segment Anything (SAM) Model.

    Args:
        backbone (keras_cv.models.Backbone): A feature extractor for the input
            images.
        prompt_encoder (keras_cv.models.SAMPromptEncoder): A Keras layer to
            compute embeddings for points, box, and mask prompt.
        mask_decoder (keras_cv.models.SAMMaskDecoder): A Keras layer to
            generate segmentation masks given the embeddings generated by the
            backbone and the prompt encoder.

    References:
        - [Segment Anything paper](https://arxiv.org/abs/2304.02643)
        - [Segment Anything GitHub](https://github.com/facebookresearch/segment-anything)

    Example:

    >>> import numpy as np
    >>> from keras_cv.src.models import ViTDetBBackbone
    >>> from keras_cv.src.models import SAMPromptEncoder
    >>> from keras_cv.src.models import SAMMaskDecoder

    Create all the components of the SAM model:

    >>> backbone = ViTDetBBackbone()
    >>> prompt_encoder = SAMPromptEncoder()
    >>> mask_decoder = SAMMaskDecoder()

    Instantiate the model:

    >>> sam = SegmentAnythingModel(
    ...     backbone=backbone,
    ...     prompt_encoder=prompt_encoder,
    ...     mask_decoder=mask_decoder
    ... )

    Define the input of the backbone. This must be a batch of images of shape
    `(1024, 1024, 3)` for the ViT backbone we are using:

    >>> image = np.ones((1, 1024, 1024, 3))

    SAM works by prompting the input images. There are three ways to prompt:

    (1) Labelled Points: Foreground points (points with label 1) are encoded
        such that the output masks generated by the mask decoder contain them
        and background points (points with label 0) are encoded such that the
        generated masks don't contain them.
    (2) Box: A box tells the model which part/crop of the image to segment.
    (3) Mask: An input mask can be used to refine the output of the mask
        decoder.

    These prompts can be mixed and matched but at least one of the prompts
    must be present. To turn off a particular prompt, simply exclude it from
    the inputs to the model.

    # TODO(ianstenbit): Remove the need for the `1` axes, and fix the box shape.
    (1) For points prompts, the expected shape is `(batch, num_points, 2)`.
        The labels must have a corresponding shape of `(batch, num_points)`.
    (2) For box prompt, the expected shape is `(batch, 1, 2, 2)`.
    (3) Similarly, mask prompts have shape `(batch, 1, H, W, 1)`.

    For example, to pass in all the prompts, do:

    >>> points = np.array([[[512., 512.], [100., 100.]]])
    >>> # For labels: 1 means foreground point, 0 means background
    >>> labels = np.array([[1., 0.]])
    >>> box = np.array([[[[384., 384.], [640., 640.]]]])
    >>> input_mask = np.ones((1, 1, 256, 256, 1))

    Prepare an input dictionary:

    >>> inputs = {
    ...     "images": image,
    ...     "points": points,
    ...     "labels": labels,
    ...     "boxes": box,
    ...     "masks": input_mask
    ... }
    ...
    >>> outputs = sam.predict(inputs)
    >>> masks, iou_pred = outputs["masks"], outputs["iou_pred"]

    The first mask in the output `masks` (i.e. `masks[:, 0, ...]`) is the best
    mask predicted by the model based on the prompts. Other `masks`
    (i.e. `masks[:, 1:, ...]`) are alternate predictions that can be used if
    they are desired over the first one.

    Now, in case of only points and box prompts, simply exclude the masks:

    >>> inputs = {
    ...     "images": image,
    ...     "points": points,
    ...     "labels": labels,
    ...     "boxes": box,
    ... }
    ...
    >>> outputs = sam.predict(inputs)
    >>> masks, iou_pred = outputs["masks"], outputs["iou_pred"]

    # TODO(ianstenbit): Remove the need for this padding.
    Another example is that only points prompts are present.
    Note that if point prompts are present but no box prompt is present, the
    points must be padded using a zero point and -1 label:

    >>> padded_points = np.concatenate(
    ...     [points, np.zeros((1, 1, 2))], axis=1
    ... )
    ...
    >>> padded_labels = np.concatenate(
    ...     [labels, -np.ones((1, 1))], axis=1
    ... )
    >>> inputs = {
    ...     "images": image,
    ...     "points": padded_points,
    ...     "labels": padded_labels,
    ... }
    ...
    >>> outputs = sam.predict(inputs)
    >>> masks, iou_pred = outputs["masks"], outputs["iou_pred"]

    Note that the segment anything model only supports inference and training
    isn't support yet. So, calling the `fit` method will fail for now.
    """  # noqa: E501

    def __init__(self, *, backbone, prompt_encoder, mask_decoder, **kwargs):
        # Get the image encoder input -- Images
        backbone_input = backbone.input

        # Define the prompt encoder inputs -- Prompts
        prompt_inputs = {
            "points": keras.Input(shape=[None, 2], name="points"),
            "labels": keras.Input(shape=[None], name="labels"),
            "boxes": keras.Input(shape=[None, 2, 2], name="boxes"),
            "masks": keras.Input(shape=[None, None, None, 1], name="masks"),
        }

        # All Inputs -- Images + Prompts
        all_inputs = {"images": backbone_input}
        all_inputs.update(prompt_inputs)

        # Build the prompt encoder
        prompt_embeddings = prompt_encoder(prompt_inputs)

        # Define the mask decoder inputs
        mask_decoder_inputs = {
            "image_embeddings": backbone.output,
            "image_pe": prompt_embeddings["dense_positional_embeddings"],
            "sparse_prompt_embeddings": prompt_embeddings["sparse_embeddings"],
            "dense_prompt_embeddings": prompt_embeddings["dense_embeddings"],
        }

        # Build the mask decoder
        outputs = mask_decoder(mask_decoder_inputs)

        super().__init__(inputs=all_inputs, outputs=outputs, **kwargs)

        self.backbone = backbone
        self.prompt_encoder = prompt_encoder
        self.mask_decoder = mask_decoder

    # TODO(ianstenbit): Do something more elegant to handle empty prompts.
    def predict_step(self, *args, **kwargs):
        if len(args) == 2:
            args = (args[0], _add_placeholder_prompts(args[-1]))
        else:
            args = (_add_placeholder_prompts(args[0]),)

        return super().predict_step(*args, **kwargs)

    def fit(self, *args, **kwargs):
        raise NotImplementedError(
            "Segment Anything Model only supports inference for now. Training"
            " the model isn't supported yet."
        )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "backbone": keras.saving.serialize_keras_object(self.backbone),
                "prompt_encoder": keras.saving.serialize_keras_object(
                    self.prompt_encoder
                ),
                "mask_decoder": keras.saving.serialize_keras_object(
                    self.mask_decoder
                ),
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        config.update(
            {
                "prompt_encoder": keras.layers.deserialize(
                    config["prompt_encoder"]
                ),
                "mask_decoder": keras.layers.deserialize(
                    config["mask_decoder"]
                ),
            }
        )
        return super().from_config(config)

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return copy.deepcopy({**backbone_presets, **sam_presets})

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return copy.deepcopy({**backbone_presets_with_weights, **sam_presets})

    @classproperty
    def backbone_presets(cls):
        """Dictionary of preset names and configurations of compatible
        backbones."""
        return copy.deepcopy(backbone_presets)


def _add_placeholder_prompts(inputs):
    """Adds placeholder prompt inputs for a call to SAM.

    Because SAM is a functional subclass model, all inputs must be specified in
    calls to the model. However, prompt inputs are all optional, so we have to
    add placeholders when they're not specified by the user.
    """
    inputs = inputs.copy()

    # Get the batch shape based on the image input
    B = ops.shape(inputs["images"])[0]

    # The type of the placeholders must match the existing inputs with respect
    # to whether or not they are tensors (as opposed to Numpy arrays).
    zeros = ops.zeros if ops.is_tensor(inputs["images"]) else np.zeros

    # Fill in missing inputs.
    if "points" not in inputs:
        inputs["points"] = zeros((B, 0, 2))
    if "labels" not in inputs:
        inputs["labels"] = zeros((B, 0))
    if "boxes" not in inputs:
        inputs["boxes"] = zeros((B, 0, 2, 2))
    if "masks" not in inputs:
        inputs["masks"] = zeros((B, 0, 256, 256, 1))

    return inputs
